import sys
import traceback
from fnmatch import fnmatch
from io import StringIO
from itertools import zip_longest
from pathlib import Path
from typing import IO, AnyStr, List, Optional, Tuple, Union

from click.testing import Result
from pynmrstar import Entry
from typer import Typer
from typer.testing import CliRunner

NOQA_E501 = "# noqa: E501"


def read_test_data(file_path: str, root_directory: Optional[str] = None) -> str:
    """
    Reads the content of a test data file from standard locations.

    This function reads the content of a test data file located at the given file path.
    If a root directory is provided, the function will consider the file path to be relative to this directory
    using the function path_in_test_data for lookup.

    Args:
        file_path (str): The path to the test data file.
        root_directory (Optional[str], optional): The root directory from which the file path is considered to be
                                                  relative.
                                                  If None, the file path is considered to be absolute.

    Returns:
        str: The content of the test data file.
    """

    file_path = (
        path_in_test_data(root_directory, file_path) if root_directory else file_path
    )
    with open(file_path) as sequence_stream:
        sequence_stream = sequence_stream.read()
    return sequence_stream


def run_and_read_pytest(args: List[str]) -> Tuple[int, str, str]:
    """
    Runs pytest with the provided arguments and captures the output.

    This function runs pytest with the given arguments and captures the return code,
    standard output, and standard error output. It temporarily redirects sys.stdout
    and sys.stderr to capture the output.

    Args:
        args (List[str]): The arguments to pass to pytest.

    Returns:
        Tuple[int, str, str]: the return code, standard output,
        and standard error output from pytest.
    """
    from pytest import main

    original_output = sys.stdout
    original_error = sys.stdout
    sys.stdout = StringIO()
    sys.stderr = StringIO()

    retcode = main(args)

    output = sys.stdout.getvalue()
    error_output = sys.stderr.getvalue()

    sys.stdout.close()
    sys.stderr.close()
    sys.stdout = original_output
    sys.stderr = original_error

    return retcode, output, error_output


def _split_test_spec(spec):
    spec_parts = spec.split("::")
    spec_parts[0] = Path(spec_parts[0])

    return spec_parts


def select_matching_tests(tests, selectors):
    results = []

    for selector in selectors:

        selector_parts = _split_test_spec(selector)
        num_selector_parts = len(selector_parts)

        if num_selector_parts == 1:
            selector = f"*::{selector_parts[0]}"
            selector_parts = _split_test_spec(selector)
            num_selector_parts = len(selector_parts)

        if num_selector_parts == 2 and selector_parts[0] == Path(""):
            selector = f"*::{selector_parts[1]}"
            selector_parts = _split_test_spec(selector)
            num_selector_parts = len(selector_parts)

        if num_selector_parts == 2 and selector_parts[1] == "":
            selector = f"{selector_parts[0]}::*"
            selector_parts = _split_test_spec(selector)
            num_selector_parts = len(selector_parts)

        selector_path_parts = selector_parts[0].parts
        num_selector_path_parts = len(selector_path_parts)

        for test in tests:

            # here we ensure we are looking for a .py file...
            if not selector_path_parts[-1].endswith("py"):
                selector_path_parts = list(selector_path_parts)
                selector_path_parts[-1] = f"{selector_path_parts[-1]}.py"
            selector_path_parts = tuple(selector_path_parts)

            test_parts = _split_test_spec(test)
            num_test_parts = len(test_parts)
            test_path_parts = test_parts[0].parts
            num_test_path_parts = len(test_path_parts)

            paths_equal = False
            if num_selector_path_parts <= num_test_path_parts:
                selector_path = selector_path_parts[-num_selector_path_parts:]
                test_path = test_path_parts[-num_selector_path_parts:]
                paths_equal = selector_path == test_path

            path = str(test_parts[0])
            path_test = str(selector_parts[0])

            if not paths_equal:
                paths_equal = fnmatch(path, path_test)

            test_names_equal = False
            if (num_selector_parts == 2) and (num_test_parts == 2):
                test_names_equal = fnmatch(test_parts[-1], selector_parts[-1])

            if paths_equal and test_names_equal:
                results.append(test)
    return results


def assert_lines_match(
    expected: str, reported: str, squash_spaces: bool = True, ignore_empty=True
):
    """
    compare two multi line strings line by line with stripping and raise an assertion if they don't match
    note: empty lines are ignoresd by default, and multiple spaces are squashed
    Args:
        expected (str): the expected string
        reported (str): the input string
        squash_spaces (bool): remove duplicate spaces before comparison

    Returns:
        None
    """
    lines_expected = expected.split("\n")
    lines_reported = reported.split("\n")

    if ignore_empty:
        lines_expected = [line for line in lines_expected if len(line.strip()) != 0]
        lines_reported = [line for line in lines_reported if len(line.strip()) != 0]

    zip_lines = zip_longest(lines_expected, lines_reported, fillvalue="")
    for i, (expected_line, reported_line) in enumerate(zip_lines, start=1):

        if squash_spaces:
            expected_line = " ".join(expected_line.split())
            reported_line = " ".join(reported_line.split())

        expected_line = expected_line.strip()
        reported_line = reported_line.strip()

        if reported_line != expected_line:

            for line_no, line in enumerate(lines_expected, start=1):
                print(f"exp|{line_no}|{line}")
            print()

            for line_no, line in enumerate(lines_reported, start=1):
                print(f"rep|{line_no}|{line}")
            print()

            print("line that caused the error:")
            print()

            print(f"exp|{i}| {expected_line}|")
            print(f"rep|{i}| {reported_line}|")

        assert reported_line == expected_line


def isolate_frame(target: str, frame_name: str) -> Optional[str]:
    """
    Extract one frame from a NEF file by name as a string
    Args:
        target (Entry): target NEF entry
        frame_name (str): name of the save frame

    Returns:
        Optional[str]: the entry or a None if not found
    """

    entry = Entry.from_string(target)
    try:
        frame = entry.get_saveframe_by_name(frame_name)
    except KeyError as e:
        msg = f"""
             the save frame {frame_name} wasn't found in the entry {entry.entry_id} the
             available_frame names are {','.join([frame.name for frame in entry.frame_list])}'
         """
        raise KeyError(msg) from e

    return str(frame)


def isolate_loop(target: str, frame_name: str, loop_category: str) -> Optional[str]:
    """
    Extract one frame from a NEF file by name as a string
    Args:
        target (Entry): target NEF entry
        frame_name (str): name of the save frame
        loop_category (str): the name of the loop

    Returns:
        Optional[str]: the frame as a string or None if it is not found
    """

    entry = Entry.from_string(target)

    try:
        frame = entry.get_saveframe_by_name(frame_name)
    except KeyError as e:
        msg = f"""
            the save frame {frame_name} wasn't found in the entry {entry.entry_id} the
            available_frame names are {','.join([frame.name for frame in entry.frame_list])}'
        """
        raise KeyError(msg) from e

    try:
        loop = str(frame.get_loop(loop_category))
    except KeyError as e:
        msg = f"""
            the loop  {loop_category} wasn't found in the frame {frame.name} in entry {entry.entry_id} the
            available_loop categories are {','.join([loop.category for loop in frame.loops])}'
        """
        raise KeyError(msg) from e

    return str(loop)


def path_in_parent_directory(root: str, target: str) -> str:
    """
    given a root and a relative path find the relative file path

    Args:
        root (str): root of the path
        target (str): the relative path from the root

    Returns:
        str: the target paths as a string
    """
    parent_path = Path(root).parent
    return str(Path(parent_path, target).absolute())


def root_path(initial_path: str):
    """given a path work up the directory structure till you find the
    directory containing the nef executable

    initial_path (str): the path to start searching from
    """

    target = Path(initial_path)

    belt_and_braces = 100  # noqa: F841 this appears to be a bug
    while (
        not (Path(target.root) == target)
        and not (target / "nef_pipelines" / "main.py").is_file()
    ):
        target = target.parent
        belt_and_braces -= 1
        if belt_and_braces < 0:
            msg = f"""\
                Error, while search for the rot of the path {initial_path} i walked up 100
                directories this looks like a bug!
            """
            raise Exception(msg)

    target /= "nef_pipelines"

    return target


# TODO: remove local we now use a hierarchical search
def path_in_test_data(root: str, file_name: str) -> str:
    """
    given a root and a file name find the relative to the file
    in the parents test_data directory

    Args:
        root (str): root of the path
        file_name (str): the name of the file

    Returns:
        str: the target paths as a string
    """

    test_data_root = root_path(root) / "tests" / "test_data"
    test_data_local = path_in_parent_directory(root, "test_data")

    if (Path(test_data_local) / file_name).is_file():
        test_data = test_data_local
    else:
        test_data = test_data_root

    # TODO: add an error on shadowing global files

    return str(Path(test_data, file_name).absolute())


def run_and_report(
    typer_app: Typer,
    args: List[str],
    input: IO[AnyStr] = None,
    expected_exit_code: int = 0,
    # separate_stderr = False # see below
) -> Result:
    """
    run a typer app in the typer test harness and report exceptions and stdout to screen if there is an error
    :param typer_app: the typer app hosting the application
    :param args: command line arguments for the app
    :param input: an input stream if required
    :param expected_exit_code: what exit code to expect if the app is expected to end with an error
    # :param separate_stderr: report STDERR in a separate stream  # see below
    :return: results object
    """

    # TODO: mix_stderr appears to be being ignored in typers/clicks test_lib, add a ticket!
    # my_mix_stderr = separate_stderr != True
    runner = CliRunner()  # mix_stderr=my_mix_stderr)

    result = runner.invoke(typer_app, args, input=input)

    if result.exit_code != expected_exit_code:
        print("\n", "-" * 40, "-stdout-", "-" * 40)
        print(result.stdout)
        if result.exception:
            print("-" * 40, "exception", "-" * 40)
            formatted = list(traceback.TracebackException(*result.exc_info).format())

            # TODO: this is a hack, I would love to have a better solution!
            if "SystemExit" in formatted[-1]:
                for i, line in enumerate(reversed(formatted)):
                    if (
                        "During handling of the above exception, another exception occurred"
                        in line
                    ):
                        break
                formatted = formatted[: -i - 2]
            print("".join(formatted))

        print("-" * 40, "-" * 9, "-" * 40)

    assert result.exit_code == expected_exit_code

    return result


def assert_frame_category_exists(
    entry: Union[Entry, str], category: str, count: int = 1, exact: bool = False
):
    """
    Asserts that the specified category exists in the given entry and
    additionally that the number of frames in that category is greater than the
    expected count if the count is specified. If you want to use an exact count
    set exact to True.

    Args:
        entry (Entry): The NEF entry to check.
        category (str): The category to look for in the entry.
        count (int, optional): The expected number of frames in the category. Defaults to 1.
        exact (bool, optional): If True, the number of frames with the category must match the expected count exactly.
    Raises:
        AssertionError: If the number of frames with the category does not match the expected count.
    """

    """
      Asserts that the specified category exists in the given entry and that the number of frames in that category
      matches the expected count [default >1].

      This function checks if the specified category exists in the given NEF entry. If the category exists, it then
      checks if the number of frames in that category is greater than or equal to the expected count. If the 'exact'
      parameter is set to True, the function checks if the number of frames in the category exactly matches
      the expected count.

      Args:
          entry (Union[Entry, str]): The NEF entry to check. This can be an Entry object or a string.
          category (str): The category to look for in the entry.
          count (int, optional): The expected number of frames in the category. Defaults to 1.
          exact (bool, optional): If True, the number of frames with the category must match the expected
          count exactly. Defaults to False.

      Raises:
          AssertionError: If the number of frames with the category does not match the expected count.
      """

    if isinstance(entry, str):
        entry = Entry.from_string(entry)
    globals_frames = entry.get_saveframes_by_category(category)
    if exact:
        assert len(globals_frames) == count
    else:
        assert len(globals_frames) >= count



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.isotope_lib import MAGNETOGYRIC_RATIO_1H
from nef_pipelines.lib.nef_frames_lib import NEF_PIPELINES_NAMESPACE
from nef_pipelines.lib.nef_lib import (
    SelectionType,
    loop_row_namespace_iter,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
    select_frames_by_name,
)
from nef_pipelines.lib.sequence_lib import translate_3_to_1
from nef_pipelines.lib.util import (
    STDIN,
    STDOUT,
    exit_error,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.modelfree import export_app

NAMESPACE = NEF_PIPELINES_NAMESPACE

MODELFREE_TYPES = "R1 R2 NOE".split()
RELAXATION_EXPERIMENT_TYPES_TO_MODELFREE_TYPES = {
    "heteronuclear_R1_relaxation": "R1",
    "heteronuclear_R2_relaxation": "R2",
    "heteronuclear_NOEs": "NOE",
}


@export_app.command()
def data(
    input_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    output_file: Path = typer.Option(
        "mfdata",
        "-o",
        "--out",
        help="file name to output to [default mfdata] for stdout use -",
        metavar="<MODELFREE-DATA>",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="force overwrite of output file if it exists and isn't empty",
    ),
    frame_selectors=typer.Argument(
        None,
        help="a list of relaxation list frames to output, [default is all], can be reepeated",
    ),
    exact: bool = typer.Option(
        False, "-e", "--exact", help="don't frame names as a wild cards"
    ),
):
    """- write a modelfree data file [alpha]"""

    if not frame_selectors:
        frame_selectors = [
            "*",
        ]
    else:
        frame_selectors = parse_comma_separated_options(frame_selectors)

    entry = read_entry_from_file_or_stdin_or_exit_error(input_file)

    entry = pipe(entry, frame_selectors, exact, Path(output_file), force)

    if entry:
        print(entry)


@dataclass
class ModelfreeRecord:
    experiment_type: str
    spectrometer_frequency: float
    value: float
    value_uncertainty: float
    flag: int = 1


def pipe(
    entry: Entry,
    frame_selectors: List[str],
    exact: bool,
    output_file: Path,
    force: bool,
):

    relaxation_list = f"{NAMESPACE}_relaxation_list"

    relaxation_lists = select_frames(entry, relaxation_list, SelectionType.CATEGORY)

    named_relaxation_lists = select_frames_by_name(
        relaxation_lists, frame_selectors, exact
    )

    _exit_if_output_file_and_no_force(output_file, force)

    relaxation_list_spectrometer_frequencies = _get_spectrometer_frequencies(
        relaxation_lists
    )

    relaxation_list_experiment_types = _get_experiment_types(relaxation_lists, entry)

    relaxation_loop_name = f"_{NAMESPACE}_relaxation"

    data_by_title = {}
    for relaxation_list in named_relaxation_lists:
        relaxation_list_name = relaxation_list.name
        relaxation_loop = (
            relaxation_list[relaxation_loop_name]
            if relaxation_loop_name in relaxation_list
            else None
        )

        for i, row in enumerate(loop_row_namespace_iter(relaxation_loop), start=1):

            title = _build_rows_title(row)

            _exit_if_title_is_too_long(title, row, i, relaxation_list, entry)

            experiment_type = relaxation_list_experiment_types[relaxation_list.name]
            spectrometer_frequency = relaxation_list_spectrometer_frequencies[
                relaxation_list_name
            ]

            modelfree_record = ModelfreeRecord(
                experiment_type,
                spectrometer_frequency,
                row.value,
                row.value_error,
            )

            values_for_title = data_by_title.setdefault(title, {})
            values_for_title[experiment_type] = modelfree_record

    table = []

    output = sys.stdout if output_file == STDOUT else output_file.open("w")
    for title, data in data_by_title.items():
        table.append(
            [
                "spin",
                title,
            ]
        )

        for name in MODELFREE_TYPES:
            datum = data[name] if name in data else None
            if not datum:
                continue
            row = [
                datum.experiment_type,
                datum.spectrometer_frequency,
                datum.value,
                datum.value_uncertainty,
                datum.flag,
            ]
            table.append(row)

        table.append([])

    print(tabulate(table, tablefmt="plain"), file=output)

    if output_file != STDOUT:
        output.close()

    return entry if output_file != STDOUT else None


def _build_rows_title(row):
    chain_code = row.chain_code_1
    sequence_code = row.sequence_code_1
    residue_name = row.residue_name_1

    residue_name_1_letter = translate_3_to_1(
        [residue_name],
    )[0]

    return f"{chain_code}_{sequence_code}_{residue_name_1_letter}"


def _exit_if_title_is_too_long(title, row, i, relaxation_list, entry):
    if len(title) > 10:
        chain_code = row.chain_code_1
        sequence_code = row.sequence_code_1
        residue_name = row.residue_name_1

        residue_name_1_letter = translate_3_to_1(
            [residue_name],
        )[0]

        msg = f"""
                    in entry {entry} in the relaxarion list {relaxation_list.name} at row {i}
                    the title derived from the first atom name is too long modelfree requires
                    less that 10 letters you have {len(title)}

                    the title was: {title}

                    it was built from

                    chain_code: {chain_code}
                    sequence_code: {sequence_code}
                    residue_name: {residue_name_1_letter} [{residue_name}]

                    your chain_code is most probably to long rename the chain with
                    nef rename chain {chain_code} A as an example

                """

        exit_error(msg)


def _get_experiment_types(relaxation_lists, entry):

    result = {}
    for relaxation_list in relaxation_lists:
        nef_experiment_type = relaxation_list.get_tag("experiment_type")[0]

        exit_if_experiment_type_isnt_recognised(
            nef_experiment_type, relaxation_list, entry
        )

        modelfree_experiment_type = RELAXATION_EXPERIMENT_TYPES_TO_MODELFREE_TYPES[
            nef_experiment_type
        ]

        result[relaxation_list.name] = modelfree_experiment_type

    return result


def exit_if_experiment_type_isnt_recognised(
    nef_experiment_type, relaxation_list, entry
):
    if nef_experiment_type not in RELAXATION_EXPERIMENT_TYPES_TO_MODELFREE_TYPES:
        msg = f"""
                in the entry {entry}

                the relaxation list {relaxation_list.name} contains an experiment type
                that the model free converter does understand it was {nef_experiment_type}
                the experiment_types this converter understands are:

                {', '.join(RELAXATION_EXPERIMENT_TYPES_TO_MODELFREE_TYPES)}

            """

        exit_error(msg)


def _exit_if_output_file_and_no_force(output_file, force):
    if output_file.exists() and not force:
        msg = f"""
            The file {output_file} exists

            use the --force option to overwrite it
        """
        exit_error(msg)


def _get_spectrometer_frequencies(relaxation_lists):
    result = {}
    for relaxation_list in relaxation_lists:
        field_strength = relaxation_list.get_tag("field_strength")[0]
        spectrometer_frequency = MAGNETOGYRIC_RATIO_1H * float(field_strength)
        result[relaxation_list.name] = spectrometer_frequency

    return result



from pathlib import Path
from typing import List

import typer

from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import ShiftList
from nef_pipelines.lib.util import STDIN
from nef_pipelines.transcoders.xeasy import import_app
from nef_pipelines.transcoders.xeasy.xeasy_lib import parse_shifts

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def shifts(
    frame_name: str = typer.Option(
        "xeasy", "-f", "--frame-name", help="a name for the frame"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type nmrview.out", metavar="<NMRVIEW-shifts>.out"
    ),
):
    """convert xeasy [flya] shift file <sparky-shifts>.txt to NEF"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    sequence = sequence_from_entry_or_exit(entry)

    pipe(entry, sequence, frame_name, file_names)


def pipe(entry, sequence, entry_name, file_names):

    xeasy_frames = []

    for file_name in file_names:

        with open(file_name) as lines:

            chain_seqid_to_type = sequence_to_residue_name_lookup(sequence)

            xeasy_shifts = parse_shifts(lines, file_name, chain_seqid_to_type)

            shift_list = ShiftList(xeasy_shifts)

            frame = shifts_to_nef_frame(shift_list, entry_name)

            xeasy_frames.append(frame)

    entry = add_frames_to_entry(entry, xeasy_frames)

    print(entry)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.csv.importers._peaks_cli import peaks

app = typer.Typer()
app.command()(peaks)


# noinspection PyUnusedLocal
def test_short_csv():

    csv_path = path_in_test_data(__file__, "ubi_hsqc_short.csv")

    result = run_and_report(app, [csv_path])

    EXPECTED = """\
        save_nef_nmr_spectrum_peaks
           _nef_nmr_spectrum.sf_category                nef_nmr_spectrum
           _nef_nmr_spectrum.sf_framecode               nef_nmr_spectrum_peaks
           _nef_nmr_spectrum.num_dimensions             2
           _nef_nmr_spectrum.chemical_shift_list        .
           _nef_nmr_spectrum.experiment_classification  .
           _nef_nmr_spectrum.experiment_type            .

           loop_
              _nef_spectrum_dimension.dimension_id
              _nef_spectrum_dimension.axis_unit
              _nef_spectrum_dimension.axis_code
              _nef_spectrum_dimension.spectrometer_frequency
              _nef_spectrum_dimension.spectral_width
              _nef_spectrum_dimension.value_first_point
              _nef_spectrum_dimension.folding
              _nef_spectrum_dimension.absolute_peak_positions
              _nef_spectrum_dimension.is_acquisition

             1   ppm   15N   60.833    8.668   123.614   none   true   false
             2   ppm   1H    600.123   1.078   9.349     none   true   true

           stop_

           loop_
              _nef_spectrum_dimension_transfer.dimension_1
              _nef_spectrum_dimension_transfer.dimension_2
              _nef_spectrum_dimension_transfer.transfer_type
              _nef_spectrum_dimension_transfer.is_indirect

             1   2   onebond   false

           stop_

           loop_
              _nef_peak.index
              _nef_peak.peak_id
              _nef_peak.chain_code_1
              _nef_peak.sequence_code_1
              _nef_peak.residue_name_1
              _nef_peak.atom_name_1
              _nef_peak.chain_code_2
              _nef_peak.sequence_code_2
              _nef_peak.residue_name_2
              _nef_peak.atom_name_2
              _nef_peak.position_1
              _nef_peak.position_uncertainty_1
              _nef_peak.position_2
              _nef_peak.position_uncertainty_2
              _nef_peak.height
              _nef_peak.height_uncertainty
              _nef_peak.volume
              _nef_peak.volume_uncertainty

             1   1   A   2   GLN   N   A   2   GLN   H   123.22   .   8.9    .   .   .   .   .
             2   2   A   3   ILE   N   A   3   ILE   H   115.34   .   8.32   .   .   .   .   .
             3   3   A   4   PHE   N   A   4   PHE   H   118.11   .   8.61   .   .   .   .   .
             4   4   A   5   VAL   N   A   5   VAL   H   121.0    .   9.3    .   .   .   .   .

           stop_

        save_

    """

    result = isolate_frame(result.stdout, "nef_nmr_spectrum_peaks")

    assert_lines_match(EXPECTED, result)



import pytest

from nef_pipelines.lib.isotope_lib import Isotope, convert_isotopes


@pytest.mark.parametrize(
    "input, expected",
    [
        ("13C", [Isotope.C13]),
        ("C13", [Isotope.C13]),
        ("12C", ["12C"]),
        (
            Isotope.C13,
            [Isotope.C13],
        ),
        ([Isotope.C13, Isotope.N15], [Isotope.C13, Isotope.N15]),
    ],
)
def test_convert_isotopes(input, expected):
    result = convert_isotopes(input)

    assert result == expected



NEF = "nef"
SF = "sf"

SHIFT_LIST_FRAME_CATEGORY = f"{NEF}_chemical_shift_list"

SPECTRUM_FRAME_CATEGORY = f"{NEF}_nmr_spectrum"
SPECTRUM_DIMENSION_LOOP_CATEGORY = f"{NEF}_spectrum_dimension"
SPECTRUM_DIMENSION_TRANSFER_LOOP_CATEGORY = (
    f"{SPECTRUM_DIMENSION_LOOP_CATEGORY}_transfer"
)
SPECTRUM_PEAK_LOOP_CATEGORY = f"{NEF}_peak"

SF_CATEGORY = f"{SF}_category"
SF_FRAMECODE = f"{SF}_framecode"
NUM_DIMENSIONS = "num_dimensions"
CHEMICAL_SHIFT_LIST = "chemical_shift_list"
EXPERIMENT_TYPE = "experiment_type"
EXPERIMENT_CLASSIFICATION = "experiment_classification"
CCPN_COMMENT = "ccpn_comment"
CCPN_MERIT = "ccpn_figure_of_merit"

NEF_NAMESPACE = "nef"
CCPN_NAMESPACE = "ccpn"
NEF_PIPELINES_NAMESPACE = "nefpls"

NEF_CHEMICAL_SHIFT_LIST = f"{NEF}_{CHEMICAL_SHIFT_LIST}"

SPECTRUM_FRAME_TAGS = (
    SF_CATEGORY,
    SF_FRAMECODE,
    NUM_DIMENSIONS,
    CHEMICAL_SHIFT_LIST,
    EXPERIMENT_CLASSIFICATION,
    EXPERIMENT_TYPE,
)

DIMENSION_ID = "dimension_id"
AXIS_UNIT = "axis_unit"
AXIS_CODE = "axis_code"

SPECTROMETER_FREQUENCY = "spectrometer_frequency"
SPECTRAL_WIDTH = "spectral_width"
VALUE_FIRST_POINT = "value_first_point"
FOLDING = "folding"
ABSOLUTE_PEAK_POSITIONS = "absolute_peak_positions"
IS_ACQUISITION = "is_acquisition"


DIMENSION_LOOP_TAGS = (
    # mandatory parameters
    DIMENSION_ID,
    AXIS_UNIT,
    AXIS_CODE,
    # # optional parameters
    SPECTROMETER_FREQUENCY,
    SPECTRAL_WIDTH,
    VALUE_FIRST_POINT,
    FOLDING,
    ABSOLUTE_PEAK_POSITIONS,
    IS_ACQUISITION,
)

DIMENSION_INDEX = "dimension_index"
DIMENSION__DIMENSION_INDEX = "dimension_{dimension_index}"
TRANSFER_TYPE = "transfer_type"
IS_INDIRECT = "is_indirect"

ONE_BOND = "onebond"
NEF_PPM = "ppm"

TRANSFER_LOOP_TAGS = [
    # mandatory parameters
    DIMENSION__DIMENSION_INDEX,
    TRANSFER_TYPE,
    # optional parameter
    IS_INDIRECT,
]

CHAIN_CODE = "chain_code"
SEQUENCE_CODE = "sequence_code"
RESIDUE_NAME = "residue_name"
ATOM_NAME = "atom_name"
POSITION = "position"
POSITION_UNCERTAINTY = "position_uncertainty"

CHAIN_CODE__DIMENSION_INDEX = f"{CHAIN_CODE}_{{dimension_index}}"
SEQUENCE_CODE__DIMENSION_INDEX = f"{SEQUENCE_CODE}_{{dimension_index}}"
RESIDUE_NAME__DIMENSION_INDEX = f"{RESIDUE_NAME}_{{dimension_index}}"
ATOM_NAME__DIMENSION_INDEX = f"{ATOM_NAME}_{{dimension_index}}"
POSITION__DIMENSION_INDEX = f"{POSITION}_{{dimension_index}}"
POSITION_UNCERTAINTY__DIMENSION_INDEX = f"{POSITION_UNCERTAINTY}_{{dimension_index}}"


PEAK_ID = "peak_id"
INDEX = "index"

HEIGHT = "height"
HEIGHT_UNCERTAINTY = "height_uncertainty"
VOLUME = "volume"
VOLUME_UNCERTAINTY = "volume_uncertainty"

PEAK_LOOP_TAGS = [
    INDEX,
    PEAK_ID,
    [
        CHAIN_CODE__DIMENSION_INDEX,
        SEQUENCE_CODE__DIMENSION_INDEX,
        RESIDUE_NAME__DIMENSION_INDEX,
        ATOM_NAME__DIMENSION_INDEX,
    ],
    [
        POSITION__DIMENSION_INDEX,
        POSITION_UNCERTAINTY__DIMENSION_INDEX,
    ],
    HEIGHT,
    HEIGHT_UNCERTAINTY,
    VOLUME,
    VOLUME_UNCERTAINTY,
]



import sys
from argparse import Namespace
from contextlib import contextmanager
from io import StringIO

import pytest

# from pandas import DataFrame
from pynmrstar import Entry, Loop

from nef_pipelines.lib.nef_lib import (  # dataframe_to_loop,; loop_to_dataframe,; NEF_CATEGORY_ATTR,
    UNUSED,
    BadNefFileException,
    create_entry_from_stdin,
    loop_row_dict_iter,
    loop_row_namespace_iter,
    read_entry_from_stdin_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
    select_frames_by_name,
)
from nef_pipelines.lib.test_lib import assert_lines_match, path_in_test_data
from nef_pipelines.lib.util import STDIN
from nef_pipelines.main import EXIT_ERROR

ITER_TEST_DATA = """\
    loop_
        _test.col_1
        _test.col_2
        _test.col_3

        a 2 4.5
        b 3 5.6

    stop_

"""

# def test_nef_to_pandas():
#
#     TEST_DATA_NEF = """
#         loop_
#           _test_loop.tag_1 _test_loop.tag_2
#           1                 2
#           3                 .
#         stop_
#     """
#
#     loop = Loop.from_string(TEST_DATA_NEF, convert_data_types=True)
#     result = loop_to_dataframe(loop)
#
#     EXPECTED_DATA_FRAME = DataFrame()
#     EXPECTED_DATA_FRAME["tag_1"] = ["1", "3"]
#     EXPECTED_DATA_FRAME["tag_2"] = ["2", "."]
#
#     assert result.equals(EXPECTED_DATA_FRAME)
#
#
# def test_pandas_to_nef():
#
#     TEST_DATA_NEF = """
#         loop_
#           _test_loop.tag_1 _test_loop.tag_2
#           1                 2
#           3                 .
#         stop_
#     """
#
#     EXPECTED_NEF = Loop.from_string(TEST_DATA_NEF)
#
#     data_frame = DataFrame()
#     data_frame["tag_1"] = ["1", "3"]
#     data_frame["tag_2"] = ["2", "."]
#
#     result = dataframe_to_loop(data_frame, category="test_loop")
#
#     assert result == EXPECTED_NEF


# def test_nef_category():
#
#     TEST_DATA_NEF = """
#         loop_
#           _test_loop.tag_1 _test_loop.tag_2
#           1                 2
#           3                 .
#         stop_
#     """
#
#     loop = Loop.from_string(TEST_DATA_NEF, convert_data_types=True)
#     frame = loop_to_dataframe(loop)
#
#     assert frame.attrs[NEF_CATEGORY_ATTR] == "test_loop"
#
#     new_loop = dataframe_to_loop(frame)
#
#     # note pynmrstar includes the leading _ in the category, I don't...!
#     assert new_loop.category == "_test_loop"
#
#     new_loop_2 = dataframe_to_loop(frame, category="wibble")
#     # note pynmrstar includes the leading _ in the category, I don't...!
#     assert new_loop_2.category == "_wibble"


def test_create_entry_from_empty_stdin(mocker):

    mocker.patch("sys.stdin", StringIO())
    result = create_entry_from_stdin()
    assert result is None


def test_create_entry_from_bad_stdin(mocker):
    mocker.patch("sys.stdin", StringIO("wibble"))

    with pytest.raises(BadNefFileException):
        result = create_entry_from_stdin()
        assert result is None


def test_create_entry_from_stdin(mocker):
    INPUT = """
        data_test
    """
    mocker.patch("sys.stdin", StringIO(INPUT))

    result = create_entry_from_stdin()

    assert result == Entry.from_scratch("test")


def test_read_or_create_entry_exit_error_on_bad_file_empty_stdin(mocker):
    mocker.patch("sys.stdin", StringIO())
    result = read_or_create_entry_exit_error_on_bad_file(STDIN)
    assert result == Entry.from_scratch("nef")


def test_read_or_create_entry_exit_error_on_bad_file_exception(mocker, tmp_path):
    mocker.patch("sys.exit")
    read_or_create_entry_exit_error_on_bad_file(tmp_path / "doesnt_exists.neff")

    sys.exit.assert_called_once_with(EXIT_ERROR)


def test_read_or_create_entry_exit_error_on_bad_file_bad_stdin(mocker):
    mocker.patch("sys.stdin", StringIO("wibble"))
    mocker.patch("sys.exit")

    read_or_create_entry_exit_error_on_bad_file(STDIN)

    sys.exit.assert_called_once_with(EXIT_ERROR)


def test_read_or_create_entry_exit_error_on_bad_file(mocker):
    INPUT = """
        data_test
    """
    mocker.patch("sys.stdin", StringIO(INPUT))

    result = create_entry_from_stdin()

    assert result == Entry.from_scratch("test")


def test_select_frames():
    TEST_DATA = """\
    data_test
        save_test_frame_1
            _test.sf_category test
            loop_
                _test.col_1
                .
            stop_
        save_


        save_test_frame_2
            _test.sf_category test
            loop_
                _test.col_1
                .
            stop_
        save_

        save_test_frame_13
            _test.sf_category test
            loop_
                _test.col_1
                .
            stop_
        save_

    """

    test = Entry.from_string(TEST_DATA)

    frames = select_frames_by_name(test, "test_frame_1")

    assert len(frames) == 1
    assert frames[0].name == "test_frame_1"

    frames = select_frames_by_name(test, ["test_frame_13"])

    assert len(frames) == 1
    assert frames[0].name == "test_frame_13"

    frames = select_frames_by_name(test, "frame_")
    assert len(frames) == 3
    names = sorted([frame.name for frame in frames])
    assert names == ["test_frame_1", "test_frame_13", "test_frame_2"]

    frames = select_frames_by_name(test, ["frame_1"], exact=False)

    assert len(frames) == 2
    names = sorted([frame.name for frame in frames])
    assert names == ["test_frame_1", "test_frame_13"]

    frames = select_frames_by_name(test, ["*frame_1*"])

    assert len(frames) == 2
    names = sorted([frame.name for frame in frames])
    assert names == ["test_frame_1", "test_frame_13"]

    frames = select_frames_by_name(test, ["frame_[1]"])

    assert len(frames) == 2
    names = sorted([frame.name for frame in frames])
    assert names == ["test_frame_1", "test_frame_13"]

    frames = select_frames_by_name(test, ["frame_[2]"])

    assert len(frames) == 1
    assert frames[0].name == "test_frame_2"


@contextmanager
def replace_stdin(target: str):
    """The provided input should be the text the user inputs. It support multiple lines for multiple inputs"""
    orig = sys.stdin
    sys.stdin = StringIO(target)
    yield
    sys.stdin = orig


def test_read_entry_stdin_or_exit_empty_stdin():
    with replace_stdin(""):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            read_entry_from_stdin_or_exit()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1


@pytest.mark.skip(reason="not currently working and deprecated")
def test_read_entry_stdin_or_exit():

    EXPECTED = """\
    data_new

    save_nef_nmr_meta_data
       _nef_nmr_meta_data.sf_category      nef_nmr_meta_data
       _nef_nmr_meta_data.sf_framecode     nef_nmr_meta_data
       _nef_nmr_meta_data.format_name      nmr_exchange_format
       _nef_nmr_meta_data.format_version   1.1
       _nef_nmr_meta_data.program_name     NEFPipelines
       _nef_nmr_meta_data.program_version  0.0.1
       _nef_nmr_meta_data.script_name      header.py
       _nef_nmr_meta_data.creation_date    2021-06-19T21:13:32.548158
       _nef_nmr_meta_data.uuid             NEFPipelines-2021-06-19T21:13:32.548158-0485797022

       loop_
          _nef_run_history.run_number
          _nef_run_history.program_name
          _nef_run_history.program_version
          _nef_run_history.script_name
          _nef_run_history.uuid

          1 NEFPipelines 1.1 header.py NEFPipelines-2021-06-19T21:13:32.548158-0485797022

       stop_

    save_

    """
    path = path_in_test_data(__file__, "header.nef")
    lines = open(path).read()

    with replace_stdin(lines):
        entry = read_entry_from_stdin_or_exit()

    assert_lines_match(str(entry), EXPECTED)


def test_loop_row_dict_iter():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"col_1": "a", "col_2": 2, "col_3": 4.5},
        {"col_1": "b", "col_2": 3, "col_3": 5.6},
    ]

    result = [row for row in loop_row_dict_iter(loop)]

    assert result == EXPECTED


def test_loop_row_dict_iter_no_convert():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"col_1": "a", "col_2": "2", "col_3": "4.5"},
        {"col_1": "b", "col_2": "3", "col_3": "5.6"},
    ]

    result = [row for row in loop_row_dict_iter(loop, convert=False)]

    assert result == EXPECTED


def test_row_dict_iter_update():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"col_1": "a", "col_2": "2_updated", "col_3": "4.5"},
        {"col_1": "b", "col_2": "3_updated", "col_3": "5.6"},
    ]

    for row in loop_row_dict_iter(loop, convert=False):
        for tag in row:
            if tag == "col_2":
                row[tag] = f"{row[tag]}_updated"

    result = [row for row in loop_row_dict_iter(loop, convert=False)]

    assert result == EXPECTED


def test_row_dict_iter_update_convert():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"col_1": "a", "col_2": "3", "col_3": "5.5"},
        {"col_1": "b", "col_2": "4", "col_3": "6.6"},
    ]

    for row in loop_row_dict_iter(loop, convert=True):
        for tag in row:
            if tag in ["col_2", "col_3"]:
                row[tag] = row[tag] + 1

    result = [row for row in loop_row_dict_iter(loop, convert=False)]

    assert result == EXPECTED


def test_row_dict_iter_delete():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"col_1": "a", "col_2": UNUSED, "col_3": "4.5"},
        {"col_1": "b", "col_2": UNUSED, "col_3": "5.6"},
    ]

    for row in loop_row_dict_iter(loop, convert=True):
        for tag in row:
            if tag == "col_2":
                del row[tag]

    result = [row for row in loop_row_dict_iter(loop, convert=False)]

    assert result == EXPECTED


def test_loop_row_namespace_iter():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        Namespace(col_1="a", col_2=2, col_3=4.5),
        Namespace(col_1="b", col_2=3, col_3=5.6),
    ]

    result = [row for row in loop_row_namespace_iter(loop)]

    assert result == EXPECTED


def test_loop_row_namespace_iter_no_convert():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        Namespace(col_1="a", col_2="2", col_3="4.5"),
        Namespace(col_1="b", col_2="3", col_3="5.6"),
    ]

    result = [row for row in loop_row_namespace_iter(loop, convert=False)]

    assert result == EXPECTED


@pytest.mark.skip(reason="not currently working")
def test_loop_row_dict_iter_attributes():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"a": "a", "b": "2", "c": "4.5"},
        {"a": "b", "b": "3", "c": "5.6"},
    ]

    for i, row in enumerate(loop_row_dict_iter(loop, convert=False)):
        assert row.col_1 == EXPECTED[i]["a"]
        assert row.col_2 == EXPECTED[i]["b"]
        assert row.col_3 == EXPECTED[i]["c"]


@pytest.mark.skip(reason="not currently working")
def test_loop_row_dict_iter_attributes_update():

    loop = Loop.from_string(ITER_TEST_DATA)

    EXPECTED = [
        {"a": "a_b", "b": "3", "c": "5.5"},
        {"a": "b_b", "b": "4", "c": "6.6"},
    ]

    for row in loop_row_dict_iter(loop, convert=True):
        row.col_1 += "_b"
        row.col_2 += 1
        row.col_3 += 1

    for i, row in enumerate(loop_row_namespace_iter(loop, convert=False)):
        assert str(row.col_1) == EXPECTED[i]["a"]
        assert str(row.col_2) == EXPECTED[i]["b"]
        assert str(row.col_3) == EXPECTED[i]["c"]



import typer
from pynmrstar import Entry

from nef_pipelines.lib.test_lib import path_in_test_data, run_and_report
from nef_pipelines.tools.frames.rename import rename

EXIT_ERROR = 1

app = typer.Typer()
app.command()(rename)

ORIGINAL_FRAME_NAMES = [
    "nef_nmr_meta_data",
    "nef_molecular_system",
    "nef_chemical_shift_list_default",
    "nef_nmr_spectrum_k_ubi_n_hsqc`1`",
    "nef_nmr_spectrum_k_ubi_hnca`1`",
    "nef_nmr_spectrum_k_ubi_hncoca`1`",
    "nef_nmr_spectrum_k_ubi_hncaco`1`",
    "nef_nmr_spectrum_k_ubi_hnco`1`",
    "nef_nmr_spectrum_k_ubi_hncacb`1`",
    "nef_nmr_spectrum_k_ubi_cbcaconh`1`",
    "nef_nmr_spectrum_mars_ubi_n_hsqc`1`",
    "ccpn_substance_1D3Z_1|Chain.None",
    "ccpn_substance_mySubstance.None",
    "ccpn_assignment",
]


# noinspection PyUnusedLocal
def test_rename_basic():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "k_ubi_hnco`1`"
    NEW_NAME = "k_ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(app, ["--in", path, SEARCHED_NAME, NEW_NAME])

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names

    assert sorted(ORIGINAL_FRAME_NAMES) != sorted(frame_names)

    NEW_FRAME_NAMES = [name for name in ORIGINAL_FRAME_NAMES if name != OLD_FRAME_ID]
    NEW_FRAME_NAMES.append(NEW_FRAME_ID)

    assert sorted(NEW_FRAME_NAMES) == sorted(frame_names)


def test_multiple_selection_fail():

    SEARCHED_NAME = "ubi"
    NEW_NAME = "k_ubi_hnco_1"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--in", path, SEARCHED_NAME, NEW_NAME], expected_exit_code=EXIT_ERROR
    )

    assert "ERROR" in result.stdout
    assert "one save frame at the same time" in result.stdout


def test_rename_short_match():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "_hnco`"
    NEW_NAME = "k_ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(app, ["--in", path, SEARCHED_NAME, NEW_NAME])

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names

    assert sorted(ORIGINAL_FRAME_NAMES) != sorted(frame_names)


def test_rename_wildcard_match():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "*hnco`"
    NEW_NAME = "k_ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(app, ["--in", path, SEARCHED_NAME, NEW_NAME])

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names

    assert sorted(ORIGINAL_FRAME_NAMES) != sorted(frame_names)


def test_rename_full_name_match():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "k_ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(app, ["--in", path, SEARCHED_NAME, NEW_NAME])

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names

    assert sorted(ORIGINAL_FRAME_NAMES) != sorted(frame_names)


def test_category_match():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "k_ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--category", CATEGORY, "--in", path, SEARCHED_NAME, NEW_NAME]
    )

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names

    assert sorted(ORIGINAL_FRAME_NAMES) != sorted(frame_names)


def test_bad_category_match():

    CATEGORY = "nef_nmr_peaklist"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "k_ubi_hnco_1"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app,
        ["--category", CATEGORY, "--in", path, SEARCHED_NAME, NEW_NAME],
        expected_exit_code=1,
    )

    assert "wasn't found in the entry" in result.stdout


def test_exact_match():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(app, ["--exact", "--in", path, SEARCHED_NAME, NEW_NAME])

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names


def test_exact_match_category():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "ubi_hnco_1"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--exact", "--category", CATEGORY, "--in", path, SEARCHED_NAME, NEW_NAME]
    )

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names


def test_exact_match_category_failure():

    CATEGORY = "ef_nmr_spectrum"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "ubi_hnco_1"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app,
        ["--exact", "--category", CATEGORY, "--in", path, SEARCHED_NAME, NEW_NAME],
        expected_exit_code=EXIT_ERROR,
    )

    assert "wasn't found in the entry" in result.stdout


def test_bad_new_name():

    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "hn co"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--in", path, SEARCHED_NAME, NEW_NAME], expected_exit_code=EXIT_ERROR
    )

    assert "frame names can't contain spaces" in result.stdout


def test_bad_exact_match():
    SEARCHED_NAME = "ef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "ubi_hnco_1"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app,
        ["--exact", "--in", path, SEARCHED_NAME, NEW_NAME],
        expected_exit_code=EXIT_ERROR,
    )

    assert "wasn't found in the entry" in result.stdout


def test_error_multiple_frames_matched():
    SEARCHED_NAME = "*"
    NEW_NAME = "ubi_hnco_1"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--in", path, SEARCHED_NAME, NEW_NAME], expected_exit_code=EXIT_ERROR
    )

    assert "multiple save frames were selected" in result.stdout


def test_error_empty_entry():
    SEARCHED_NAME = "*"
    NEW_NAME = "ubi_hnco_1"

    path = path_in_test_data(__file__, "empty.nef")
    result = run_and_report(
        app, ["--in", path, SEARCHED_NAME, NEW_NAME], expected_exit_code=EXIT_ERROR
    )

    assert "there were no frames in the entry" in result.stdout


def test_error_no_frames_selected():
    SEARCHED_NAME = "wibble"
    NEW_NAME = "k_ubi_hnco_1"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--in", path, SEARCHED_NAME, NEW_NAME], expected_exit_code=EXIT_ERROR
    )

    assert "wasn't found in the entry" in result.stdout


def test_rename_to_existing_fail():
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "k_ubi_hncaco`1`"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--in", path, SEARCHED_NAME, NEW_NAME], expected_exit_code=EXIT_ERROR
    )

    assert "a frame with the name" in result.stdout
    assert NEW_NAME in result.stdout
    assert "already exists in entry" in result.stdout


def test_rename_to_existing_force():
    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "nef_nmr_spectrum_k_ubi_hnco`1`"
    NEW_NAME = "k_ubi_hncaco`1`"
    NEW_FRAME_ID = f"{CATEGORY}_{NEW_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(app, ["--in", path, "--force", SEARCHED_NAME, NEW_NAME])

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert SEARCHED_NAME not in frame_names


def test_rename_category():

    CATEGORY = "nef_nmr_spectrum"
    SEARCHED_NAME = "k_ubi_hnco`1`"
    NEW_NAME = "nef_test"
    OLD_FRAME_ID = f"{CATEGORY}_{SEARCHED_NAME}"
    NEW_FRAME_ID = f"{NEW_NAME}_{SEARCHED_NAME}"

    path = path_in_test_data(__file__, "ubiquitin_short.nef")
    result = run_and_report(
        app, ["--in", path, "--rename-category", SEARCHED_NAME, NEW_NAME]
    )

    entry = Entry.from_string(result.stdout)

    frame_names = list(entry.frame_dict.keys())
    assert NEW_FRAME_ID in frame_names
    assert OLD_FRAME_ID not in frame_names

    assert sorted(ORIGINAL_FRAME_NAMES) != sorted(frame_names)

    NEW_FRAME_NAMES = [name for name in ORIGINAL_FRAME_NAMES if name != OLD_FRAME_ID]
    NEW_FRAME_NAMES.append(NEW_FRAME_ID)

    assert sorted(NEW_FRAME_NAMES) == sorted(frame_names)

    assert ["sf_category", NEW_NAME] in entry.frame_dict[NEW_FRAME_ID].tags



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.shifty.exporters.shifts import shifts

runner = CliRunner()
app = typer.Typer()
app.command()(shifts)

EXPECTED = """\
#NUM  AA     HA     HN      N15      CA      CB       CO
1     A   5.320  0.000  115.740  53.270  20.340  181.080
2     G   0.000  7.310  117.940  46.770   0.000  176.920
3     V   5.020  8.820  116.850  71.710  33.150  172.980
"""


# noinspection PyUnusedLocal
def test_3ab():

    input = read_test_data("test_agv.neff", __file__)
    result = run_and_report(app, ["-"], input=input)

    assert_lines_match(EXPECTED, result.stdout)



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xcamshift.exporters.shifts import shifts

runner = CliRunner()
app = typer.Typer()
app.command()(shifts)

EXPECTED = """\
assign  (  segid  AAAA  and  resid  1  and  resn  ALA  and  name  N    )  115.740
assign  (  segid  AAAA  and  resid  1  and  resn  ALA  and  name  C    )  181.080
assign  (  segid  AAAA  and  resid  1  and  resn  ALA  and  name  CA   )   53.270
assign  (  segid  AAAA  and  resid  1  and  resn  ALA  and  name  CB   )   20.340
assign  (  segid  AAAA  and  resid  1  and  resn  ALA  and  name  HA   )    5.320
assign  (  segid  AAAA  and  resid  2  and  resn  GLY  and  name  HN   )    7.310
assign  (  segid  AAAA  and  resid  2  and  resn  GLY  and  name  N    )  117.940
assign  (  segid  AAAA  and  resid  2  and  resn  GLY  and  name  C    )  176.920
assign  (  segid  AAAA  and  resid  2  and  resn  GLY  and  name  CA   )   46.770
assign  (  segid  AAAA  and  resid  2  and  resn  GLY  and  name  HA2  )    4.890
assign  (  segid  AAAA  and  resid  3  and  resn  VAL  and  name  HN   )    8.820
assign  (  segid  AAAA  and  resid  3  and  resn  VAL  and  name  N    )  116.850
assign  (  segid  AAAA  and  resid  3  and  resn  VAL  and  name  C    )  172.980
assign  (  segid  AAAA  and  resid  3  and  resn  VAL  and  name  CA   )   71.710
assign  (  segid  AAAA  and  resid  3  and  resn  VAL  and  name  CB   )   33.150
assign  (  segid  AAAA  and  resid  3  and  resn  VAL  and  name  HA   )    5.020
"""


# noinspection PyUnusedLocal
def test_3ab():

    input = read_test_data("test_agv.neff", __file__)
    result = run_and_report(app, ["-"], input=input)

    assert_lines_match(EXPECTED, result.stdout)



from dataclasses import replace
from enum import auto
from typing import Tuple

from strenum import LowercaseStrEnum

from nef_pipelines.lib.structures import AtomLabel, NewPeak


# this is just a place holder
class NamingScheme(LowercaseStrEnum):
    XPLOR = auto()
    IUPAC = auto()


TranslationScheme = Tuple[NamingScheme, NamingScheme]

XPLOR_IUPAC = NamingScheme.XPLOR, NamingScheme.IUPAC
IUPAC_XPLOR = NamingScheme.IUPAC, NamingScheme.XPLOR
TRANSLATIONS = {
    XPLOR_IUPAC: {
        "HN": "H",
    },
    IUPAC_XPLOR: {
        "H": "HN",
    },
}


def translate_atom_label(
    atom: AtomLabel, translation_scheme: TranslationScheme = XPLOR_IUPAC
) -> AtomLabel:
    """
    Translate an atom name form on sheme to another e.g xplor to iupac
    :param atom: the atom label to translate
    :param translation_scheme: the translation scheme which is a tuple of NamingSchemes defining a translation from the
                                first to the second
    :return: the renamed atom label which is a new instance
    """
    if translation_scheme in TRANSLATIONS:
        translation = TRANSLATIONS[translation_scheme]
    else:
        translation = translation_scheme

    atom_name = atom.atom_name
    if atom_name in translation:
        translated_atom_name = translation[atom_name]
        atom = replace(atom, atom_name=translated_atom_name)
    return atom


def translate_new_peak(
    peak: NewPeak, translation_scheme: TranslationScheme = XPLOR_IUPAC
) -> NewPeak:
    """
    Translate the atom names in a new peak to use a different naming convention according the the translation schema
    :param peak: the peak whose labels need renaming
    :param translation_scheme: the translation scheme which is a tuple of NamingSchemes defining a translation from the
                                first to the second
    :return: the peak with renamed atom labels this is a new instance
    """
    new_shifts = []
    for shift in peak.shifts:
        atom = translate_atom_label(shift.atom, translation_scheme)
        new_shifts.append(replace(shift, atom=atom))

    result = replace(peak, shifts=new_shifts)

    return result



import re
import sys
from pathlib import Path
from typing import Dict, List

import typer
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    molecular_system_from_entry_or_exit,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (
    make_chunked_sequence_1let,
    sequence_to_chains,
    sequences_from_frames,
    translate_3_to_1,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import (
    STDIN,
    STDOUT,
    exit_error,
    exit_if_file_has_bytes_and_no_force,
)
from nef_pipelines.transcoders.fasta import export_app

# TODO: we should be able to output *s on the ends of sequences and comments
# TODO: we should be able to select chains by fnmatch

CHAIN_CODES_HELP = """
    chains to export, to add multiple chains use repeated calls, to select all chains use * [default: 'A']
"""


# noinspection PyUnusedLocal
@export_app.command()
def sequence(
    chain_codes: List[str] = typer.Option(
        [],
        "-c",
        "--chain_code",
        help=CHAIN_CODES_HELP,
        metavar="<CHAIN-CODE>",
    ),
    in_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    output_file: str = typer.Argument(
        None,
        help="file name to output to [default <entry_id>.fasta] for stdout use -",
        metavar="<FASTA-SEQUENCE-FILE>",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="force overwrite of output file if it exists and isn't empty",
    ),
):
    """- convert nef sequence to fasta"""

    chain_codes = ["*"] if len(chain_codes) == 0 else chain_codes

    entry = read_entry_from_file_or_stdin_or_exit_error(in_file)

    output_file = f"{entry.entry_id}.fasta" if output_file is None else output_file
    output_file = STDOUT if output_file == "-" else output_file

    entry = pipe(entry, chain_codes, Path(output_file), force)

    if entry:
        print(entry)


def pipe(entry: Entry, chain_codes: List[str], output_file: Path, force: bool):

    fasta_records = fasta_records_from_entry(entry, chain_codes)

    exit_if_file_has_bytes_and_no_force(output_file, force)

    # TODO: move to utility function and use in all outputs
    try:
        file_h = sys.stdout if output_file == STDOUT else open(output_file, "w")
    except Exception as e:
        msg = (
            f"Error opening output file {output_file} for writing fasa file because {e}"
        )
        exit_error(msg, e)

    # TODO we ought to be able to output to multiple files with a filename template
    records = []
    for record in fasta_records.values():
        records.append("\n".join(record))
    print("\n\n".join(records), file=file_h)

    if output_file != STDOUT:
        file_h.close()

    return entry if output_file != STDOUT else None


def fasta_records_from_entry(entry, chain_codes):

    molecular_system = molecular_system_from_entry_or_exit(entry)

    residues = sequences_from_frames(molecular_system)

    fasta_records = nef_to_fasta_records(residues, chain_codes, entry.entry_id)

    return fasta_records


def nef_to_fasta_records(
    residues: List[SequenceResidue], target_chain_codes: List[str], entry_id: str
) -> Dict[str, List[str]]:

    all_chain_codes = sequence_to_chains(residues)

    chain_codes = []
    for target_chain_code in target_chain_codes:
        for chain_code in all_chain_codes:
            if target_chain_code == chain_code:
                chain_codes.append(chain_code)
                continue
            if target_chain_code == "*":
                chain_codes.append(chain_code)
                continue

    residues_by_chain = {}
    for chain_code in chain_codes:
        residues_by_chain[chain_code] = [
            residue for residue in residues if residue.chain_code == chain_code
        ]

    chain_starts = {}
    for chain_code, residues in residues_by_chain.items():
        chain_starts[chain_code] = residues[0].sequence_code

    residue_sequences = {}
    for chain_code in chain_codes:
        residue_sequences[chain_code] = [
            residue.residue_name for residue in residues_by_chain[chain_code]
        ]

    residue_sequences = {
        chain: translate_3_to_1(residue_sequence)
        for chain, residue_sequence in residue_sequences.items()
    }
    residue_sequences = {
        chain: make_chunked_sequence_1let(residue_sequence)
        for chain, residue_sequence in residue_sequences.items()
    }

    result = {}
    for chain_code, residue_sequence in residue_sequences.items():
        entry_id = re.sub(r"\s+", "", entry_id)
        result[chain_code] = [
            f">{entry_id} NEFPLS | CHAIN: {chain_code} | START: {chain_starts[chain_code]}",
            "\n".join(residue_sequence),
        ]

    return result



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrpipe.importers.peaks import peaks

HEADER = read_test_data("test_header_entry.txt", __file__)

METADATA_NMRPIPE = "nef_nmr_meta_data"


app = typer.Typer()
app.command()(peaks)


# noinspection PyUnusedLocal
def test_peaks():
    EXPECTED = read_test_data("gb3_assigned_trunc_expected.nef", __file__)

    path = path_in_test_data(__file__, "gb3_assigned_trunc.tab")
    result = run_and_report(app, [path], input=HEADER)

    peaks_result = isolate_frame(result.stdout, "nef_nmr_spectrum_gb3_assigned_trunc")

    assert_lines_match(EXPECTED, peaks_result)



import os
from pathlib import Path
from textwrap import dedent
from typing import List

import typer

from nef_pipelines.lib.test_lib import run_and_read_pytest, select_matching_tests
from nef_pipelines.lib.util import exit_error
from nef_pipelines.nef_app import app

TARGET_HELP = """
    targets are selected as path::name, globs can be used,
    missing paths and names are replaced by * eg path:: is equivalent to path::*
    note: filenames do not require a trailing .py
    """
TARGET_HELP = dedent(TARGET_HELP)


# TODO add verbosity control
# TODO add flag to enable warnings


@app.command()
def test(
    warnings: bool = typer.Option(
        False, "-w", "--warnings", help="include all warnings"
    ),
    # native_tracebacks: bool = typer.Option(False, '-n', '--native-tracebacks', help='use standard python tracebacks'),
    targets: List[str] = typer.Argument(None, help=TARGET_HELP),
):
    """- run the test suite"""

    from pytest import main

    dir_path = Path(os.path.dirname(os.path.realpath(__file__))).parent

    root_path = str(dir_path.parent / "nef_pipelines" / "tests")

    os.chdir(Path(root_path).absolute())

    tests = _find_pytest_commands(root_path, targets)

    if not targets or (targets and len(tests) != 0):
        command = ["-vvv", "--full-trace", *tests]

        if not warnings:
            command = ["--disable-warnings", *command]
        # if native_tracebacks:
        #     command = ['--tb=native', *command]
        main(command)


def _find_pytest_root(root_path):

    stdout = _run_pytest_and_exit_error(["--fixtures", root_path])

    result = None

    for line in stdout.split("\n"):

        if line.startswith("rootdir:"):
            spaced_fields = line.split()
            fields = []

            for field in spaced_fields:
                fields.extend(field.split(","))
            result = fields[1]
            break

    return result


def _find_pytest_commands(root_path, targets):
    if not targets:
        targets = "*"

    stdout = _run_pytest_and_exit_error(["--collect-only", "-qq", root_path])

    pytest_root = _find_pytest_root(root_path)

    if not pytest_root:
        exit_error(f"couldn't find pytest root directory starting from: {root_path}")

    tests = stdout.split("\n")[:-3]

    result = [
        Path(pytest_root) / test_path
        for test_path in select_matching_tests(tests, targets)
    ]

    return result


def _run_pytest_and_exit_error(commands):
    ret_code, stdout, stderr = run_and_read_pytest(commands)

    if ret_code != 0:
        msg = f"""
        running pytest failed:

        return code: {ret_code}
        __________________________________________________________________stdout_______________________________________________________________
        {stdout}
        _______________________________________________________________________________________________________________________________________
        __________________________________________________________________stderr_______________________________________________________________
        {stderr}
        _______________________________________________________________________________________________________________________________________
        """

        exit_error(msg)

    return stdout


def _exit_if_stderr(stderr):
    if stderr.strip():
        _exit_pytest_error(stderr)


def _exit_pytest_error(output):
    msg = f"""
               couldn't collect tests because of an error

               ----------------------- pytest errors -----------------------
               {output}
               ----------------------- pytest errors -----------------------
            """
    msg = dedent(msg)
    msg = f"{msg}"
    exit_error(msg)



import typer
from pytest import fixture

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.csv.importers._rdcs_cli import rdcs

app = typer.Typer()
app.command()(rdcs)


@fixture
def INPUT_3A_AB_NEF():
    return read_test_data("3a_ab.neff", __file__)


# noinspection PyUnusedLocal
def test_short_csv(INPUT_3A_AB_NEF):
    csv_path = path_in_test_data(__file__, "short.csv")

    args = [csv_path, "--chain-code", "AAAA"]
    result = run_and_report(app, args, input=INPUT_3A_AB_NEF)

    EXPECTED = """\
        save_nef_rdc_restraint_list_rdcs
           _nef_rdc_restraint_list.sf_category           nef_rdc_restraint_list
           _nef_rdc_restraint_list.sf_framecode          nef_rdc_restraint_list_rdcs
           _nef_rdc_restraint_list.restraint_origin      .
           _nef_rdc_restraint_list.tensor_magnitude      .
           _nef_rdc_restraint_list.tensor_rhombicity     .
           _nef_rdc_restraint_list.tensor_chain_code     .
           _nef_rdc_restraint_list.tensor_sequence_code  .
           _nef_rdc_restraint_list.tensor_residue_name   .

           loop_
              _nef_rdc_restraint.index
              _nef_rdc_restraint.restraint_id
              _nef_rdc_restraint.restraint_combination_id
              _nef_rdc_restraint.chain_code_1
              _nef_rdc_restraint.sequence_code_1
              _nef_rdc_restraint.residue_name_1
              _nef_rdc_restraint.atom_name_1
              _nef_rdc_restraint.chain_code_2
              _nef_rdc_restraint.sequence_code_2
              _nef_rdc_restraint.residue_name_2
              _nef_rdc_restraint.atom_name_2
              _nef_rdc_restraint.weight
              _nef_rdc_restraint.target_value
              _nef_rdc_restraint.target_value_uncertainty
              _nef_rdc_restraint.lower_linear_limit
              _nef_rdc_restraint.lower_limit
              _nef_rdc_restraint.upper_limit
              _nef_rdc_restraint.upper_linear_limit
              _nef_rdc_restraint.scale
              _nef_rdc_restraint.distance_dependent

             0   0   .   AAAA   1   ALA   H   AAAA   1   ALA   N   1.0   1.3   .   .   .   .   .   1.0   .
             1   1   .   AAAA   2   ALA   H   AAAA   2   ALA   N   1.0   4.6   .   .   .   .   .   1.0   .
             2   2   .   AAAA   3   ALA   H   AAAA   3   ALA   N   1.0   2.4   .   .   .   .   .   1.0   .

           stop_

        save_

    """

    print(result.stdout)
    result = isolate_frame(result.stdout, "nef_rdc_restraint_list_rdcs")

    assert_lines_match(EXPECTED, result)


def test_short_complete_csv(INPUT_3A_AB_NEF):
    csv_path = path_in_test_data(__file__, "short_complete.csv")

    args = [csv_path]
    result = run_and_report(app, args, input=INPUT_3A_AB_NEF)

    EXPECTED = """\
        save_nef_rdc_restraint_list_rdcs
           _nef_rdc_restraint_list.sf_category           nef_rdc_restraint_list
           _nef_rdc_restraint_list.sf_framecode          nef_rdc_restraint_list_rdcs
           _nef_rdc_restraint_list.restraint_origin      .
           _nef_rdc_restraint_list.tensor_magnitude      .
           _nef_rdc_restraint_list.tensor_rhombicity     .
           _nef_rdc_restraint_list.tensor_chain_code     .
           _nef_rdc_restraint_list.tensor_sequence_code  .
           _nef_rdc_restraint_list.tensor_residue_name   .

           loop_
              _nef_rdc_restraint.index
              _nef_rdc_restraint.restraint_id
              _nef_rdc_restraint.restraint_combination_id
              _nef_rdc_restraint.chain_code_1
              _nef_rdc_restraint.sequence_code_1
              _nef_rdc_restraint.residue_name_1
              _nef_rdc_restraint.atom_name_1
              _nef_rdc_restraint.chain_code_2
              _nef_rdc_restraint.sequence_code_2
              _nef_rdc_restraint.residue_name_2
              _nef_rdc_restraint.atom_name_2
              _nef_rdc_restraint.weight
              _nef_rdc_restraint.target_value
              _nef_rdc_restraint.target_value_uncertainty
              _nef_rdc_restraint.lower_linear_limit
              _nef_rdc_restraint.lower_limit
              _nef_rdc_restraint.upper_limit
              _nef_rdc_restraint.upper_linear_limit
              _nef_rdc_restraint.scale
              _nef_rdc_restraint.distance_dependent

             0   0   .   AAAA   1   ALA   HA     AAAA   1   ALA   HN   1.0   1.3   1.0   .   .   .   .   1.0   .
             1   1   .   AAAA   1   ALA   HB     AAAA   1   ALA   HN   1.0   4.6   2.0   .   .   .   .   1.0   .
             2   2   .   AAAA   2   ALA   HG3#   AAAA   2   ALA   HN   1.0   2.4   3.0   .   .   .   .   1.0   .
             3   3   .   AAAA   3   ALA   HA     AAAA   3   ALA   HN   1.0   6.7   4.0   .   .   .   .   1.0   .

           stop_

        save_

    """

    result = isolate_frame(result.stdout, "nef_rdc_restraint_list_rdcs")

    assert_lines_match(EXPECTED, result)



import typer

from nef_pipelines import nef_app

fit_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        fit_app, name="fit", help="- carry out fitting operations [alpha]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.fit.exponential  # noqa: F401
    import nef_pipelines.tools.fit.mean  # noqa: F401
    import nef_pipelines.tools.fit.ratio  # noqa: F401
    import nef_pipelines.tools.fit.t1noe  # noqa: F401



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.fit.mean import mean

runner = CliRunner()
app = typer.Typer()
app.command()(mean)


EXPECTED_DATA_MEAN = """
   loop_
      _nefpls_relaxation.index
      _nefpls_relaxation.data_id
      _nefpls_relaxation.data_combination_id
      _nefpls_relaxation.value
      _nefpls_relaxation.value_error
      _nefpls_relaxation.chain_code_1
      _nefpls_relaxation.sequence_code_1
      _nefpls_relaxation.residue_name_1
      _nefpls_relaxation.atom_name_1
      _nefpls_relaxation.chain_code_2
      _nefpls_relaxation.sequence_code_2
      _nefpls_relaxation.residue_name_2
      _nefpls_relaxation.atom_name_2

     1   1   .   1.000000  0.244949   A   18   LYS   H   A   18   LYS   N

   stop_
"""


def test_mean_data_single():
    """Test exponential fitting with test_1_exponential.nef test data."""

    test_data = open(path_in_test_data(__file__, "test_1_mean.nef")).read()

    # Test that function exits with error when no noise level provided
    result = run_and_report(app, ["MEAN"], input=test_data)

    mean_loop = isolate_loop(
        result.stdout, "nefpls_relaxation_list_mean", "nefpls_relaxation"
    )
    assert_lines_match(EXPECTED_DATA_MEAN, mean_loop)



import typer
from fyeah import f

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.talos.importers.secondary_structure import (
    secondary_structure,
)

app = typer.Typer()
app.command()(secondary_structure)

EXPECTED = """
save_nefpls_secondary_structure_A_talos                                                                     # noqa: E501
   _nefpls_secondary_structure.sf_category   nefpls_secondary_structure
   _nefpls_secondary_structure.sf_framecode  nefpls_secondary_structure_A_talos
   _nefpls_secondary_structure.method        talos
   _nefpls_secondary_structure.version       .

   loop_
      _nefpls_secondary_structure.index
      _nefpls_secondary_structure.chain_code
      _nefpls_secondary_structure.sequence_code
      _nefpls_secondary_structure.residue_name
      _nefpls_secondary_structure.secondary_structure
      _nefpls_secondary_structure.merit
      _nefpls_secondary_structure.comment

     1   A   1   MET   coil         0.0     .
     2   A   2   GLN   beta_sheet   0.5     .
     3   A   3   ILE   beta_sheet   0.94    .
     4   A   4   PHE   beta_sheet   0.94    .

   stop_

save_
""".replace(
    NOQA_E501, ""
)


def test_ss_4():

    pred_4_path = path_in_test_data(__file__, "predSS_4.tab")
    pred_4_nef_path = path_in_test_data(__file__, "pred_4_seq.nef")

    STREAM = read_test_data(pred_4_nef_path)

    result = run_and_report(
        app,
        [
            pred_4_path,
        ],
        input=STREAM,
    )

    phi_psi = isolate_frame(result.stdout, "nefpls_secondary_structure_A_talos")

    PATCHED_EXPECTED = f(EXPECTED)
    assert_lines_match(PATCHED_EXPECTED, phi_psi)


EXPECTED_FIRST_RESID_2 = """
save_nefpls_secondary_structure_A_talos                                                                     # noqa: E501
   _nefpls_secondary_structure.sf_category   nefpls_secondary_structure
   _nefpls_secondary_structure.sf_framecode  nefpls_secondary_structure_A_talos
   _nefpls_secondary_structure.method        talos
   _nefpls_secondary_structure.version       .

   loop_
      _nefpls_secondary_structure.index
      _nefpls_secondary_structure.chain_code
      _nefpls_secondary_structure.sequence_code
      _nefpls_secondary_structure.residue_name
      _nefpls_secondary_structure.secondary_structure
      _nefpls_secondary_structure.merit
      _nefpls_secondary_structure.comment

     1   A   2   MET   coil         0.0     .
     2   A   3   GLN   beta_sheet   0.5     .
     3   A   4   ILE   beta_sheet   0.94    .
     4   A   5   PHE   beta_sheet   0.94    .

   stop_

save_
""".replace(
    NOQA_E501, ""
)


def test_ss_4_first_resid_2():

    pred_4_path = path_in_test_data(__file__, "predSS_4_first_resid_2.tab")
    pred_4_nef_path = path_in_test_data(__file__, "pred_4_seq_first_resid_2.nef")

    with open(pred_4_nef_path) as fh:
        STREAM = fh.read()

    result = run_and_report(
        app,
        [
            pred_4_path,
        ],
        input=STREAM,
    )

    phi_psi = isolate_frame(result.stdout, "nefpls_secondary_structure_A_talos")

    PATCHED_EXPECTED = f(EXPECTED_FIRST_RESID_2)
    assert_lines_match(PATCHED_EXPECTED, phi_psi)



import sys
from copy import deepcopy
from dataclasses import dataclass, replace
from enum import auto
from pathlib import Path
from typing import Dict, List

import typer
from fyeah import f
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum
from tabulate import tabulate
from uncertainties import ufloat

from nef_pipelines.lib.nef_frames_lib import (
    EXPERIMENT_CLASSIFICATION,
    EXPERIMENT_TYPE,
    NEF_PIPELINES_NAMESPACE,
)
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    SelectionType,
    create_nef_save_frame,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks, peaks_to_frame
from nef_pipelines.lib.sequence_lib import sequences_from_frames
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.spectra_lib import (
    EXPERIMENT_CLASSIFICATION_TO_SYNONYM,
    EXPERIMENT_INFO,
    FAKE_SPECTROMETER_FREQUENCY_600,
    SPECTRUM_TYPE_TO_CLASSIFICATION,
    ExperimentType,
)
from nef_pipelines.lib.util import exit_error, parse_comma_separated_options
from nef_pipelines.tools.simulate import simulate_app
from nef_pipelines.tools.simulate.peaks import _make_peak_list

SPECTRUM_TYPE_NHSQC = ExperimentType.N_HSQC


class AminoAcid(StrEnum):
    ALA = auto()
    ARG = auto()
    ASN = auto()
    ASP = auto()
    CYS = auto()
    GLN = auto()
    GLU = auto()
    GLY = auto()
    HIS = auto()
    ILE = auto()
    LEU = auto()
    LYS = auto()
    MET = auto()
    PHE = auto()
    PRO = auto()
    SER = auto()
    THR = auto()
    TRP = auto()
    TYR = auto()
    VAL = auto()
    REF = auto()
    ANY = "*"


ALL_AMINO_ACID_SET = {*AminoAcid.__members__.values()}
ALL_AMINO_ACID_SET.remove(AminoAcid.REF)
ALL_AMINO_ACID_SET.remove(AminoAcid.ANY)
ALL_AMINO_ACID_SET.remove(AminoAcid.PRO)

FULLY_SCRAMBLED_SET = {
    AminoAcid.ASP,
    AminoAcid.GLU,
    AminoAcid.SER,
}

# residues which are not global scramblers
USABLE_UNLABELLING_SET = ALL_AMINO_ACID_SET - FULLY_SCRAMBLED_SET

OBVIOUS_SHIFT_SET = {AminoAcid.ALA, AminoAcid.GLY, AminoAcid.SER, AminoAcid.THR}


SINGLY_SCRAMBLED_SET = {  # GB1   APRATAXIN
    AminoAcid.ALA,  # VAL    VAL
    AminoAcid.GLY,  # TRP    CYS  SER  TRP
    AminoAcid.PHE,  # TYR    TYR
    AminoAcid.TYR,  # PHE                                          ! NOT REALLY UNLABELLED
}

# note residues in brackets are not really scrambled that much
MULTIPLY_SCRAMBLED_SET = {
    # GB1                   APRATAXIN
    AminoAcid.ILE,  # LEU  VAL              LEU  VAL  [ALA]
    AminoAcid.LEU,  # ILE  [VAL]            ILE  VAL  [ALA]
    AminoAcid.THR,  # GLY  TRP              GLY  CYS  SER    [TRP]
    AminoAcid.TRP,  # TYR  PHE              TYR  PHE                ! NOT REALLY UNLABELLED
    AminoAcid.VAL,  # ILE  LEU  [ALA]       ILE  LEU  [ALA]
}

CLEAN_SET = USABLE_UNLABELLING_SET - MULTIPLY_SCRAMBLED_SET - SINGLY_SCRAMBLED_SET
SEMI_CLEAN_SET = USABLE_UNLABELLING_SET - SINGLY_SCRAMBLED_SET

NO_OBV_CLEAN_SET = CLEAN_SET - OBVIOUS_SHIFT_SET
NO_OBV_SEMI_CLEAN_SET = SEMI_CLEAN_SET - OBVIOUS_SHIFT_SET


LABELING_SETS = {
    "ALL": ALL_AMINO_ACID_SET,
    "USEABLE": USABLE_UNLABELLING_SET,
    "CLEAN": CLEAN_SET,
    "SEMI_CLEAN": SEMI_CLEAN_SET,
    "NO_OBV_CLEAN": NO_OBV_CLEAN_SET,
    "NO_OBV_SEMI_CLEAN": NO_OBV_SEMI_CLEAN_SET,
    "MULTIPLY_SCRAMBLED": MULTIPLY_SCRAMBLED_SET,
    "SINGLY_SCRAMBLED": SINGLY_SCRAMBLED_SET,
    "REMOVE_OBVIOUS": OBVIOUS_SHIFT_SET,
}

LABELING_SET_HELP = {
    "ALL": "All amino acids",
    "USEABLE": "All amino acids excluding those that are uniformly scrambled",
    "CLEAN": "Amino acids that clearly have no scrambling",
    "SEMI_CLEAN": "Amino acids that have no scrambling or a single scambling partner",
    "NO_OBV_CLEAN": "CLEAN set with amino acids with obvious CB shifts removed",
    "NO_OBV_SEMI_CLEAN": "SEMI_CLEAN set with amino acids with obvious CB shifts removed",
    "MULTIPLY_SCRAMBLED": "Amino acids that are scrambled to mutiple partners",
    "SINGLY_SCRAMBLED": "Amino acids that are scrambled to single partners",
    "REMOVE_OBVIOUS": "Causes amino acids with obvious HN, Cα or Cβ chemical shifts to be removed, remove:",
}

ALL_AMINO_ACIDS = [aa.value for aa in AminoAcid if aa.value != "*"]
ALL_AMINO_ACIDS_UPPER = [aa.upper() for aa in ALL_AMINO_ACIDS]


UNLABELLED_SPECTRUM_NAME_TEMPLATE = "synthetic_{spectrum}_unlabelled_{residues}"

UNLABELLED_AMINO_ACIDS_AND_SETS_HELP = """\
    the amino acids or sets of amino acids to use for the unlabelling, to list the available amino acids and sets
    use the --list option. Note the sets are additive except for REMOVE_OBVIOUS_CBETAS which is subtractive.
    The default is NO_CB_CLEAN
"""


FRAME_SELECTORS_HELP = """\
    the names of the frames to use to create the unlabelling, this can be a comma separated list of frame names or
    the option can be called called multiple times. Wild cards are allowed. If no match is found wild cards are checked
    as well  unless --exact is set. Frames can be from any category if --residue-types is used otherwise they should be
    nmr_spetra [aka peak lists]. If no frames are selected and --residue-types is set the default chemical shift list is
    used otherwise all peak lists are used.
"""

NAME_TEMPLATE_HELP = """
    the name template for the new peak lists the placeholders {spectrum} and {residues} will get replaced with the
    spectrum name and the unlabelled residues respectively
"""

RESIDUE_TYPES_HELP = """\
output a residue type list rather than simulated peaks, each residue type will be an anonymous residue with the same
sequence code as the assigned reside i.e.
    A 74 VAL H -> @- @74 '' H.
The format of the residue type list can be printed using the --residue-types-format option
"""

DISPLAY_RESIDUE_TYPES_FORMAT_HELP = (
    "display the NEF format used by the --residue-types option"
)

RESIDUE_TYPES_FORMAT = """
    The Format of the residue type list is

    save_nefpls_residue_types_default
       _nefpls.sf_category nefpls_residue_types
       _nefpls.sf_framecode nefpls_residue_types_default

       loop_
            _nefpls_residue_type.index
            _nefpls_residue_type.chain_code
            _nefpls_residue_type.sequence_code
            _nefpls_residue_type.residue_type
            _nefpls_residue_type.probability

            1    @-     @1      ALA    0.3
            2    @-     @1      ARG    0.5
            3    @-     @2-1    ASN    .
            4    @-     @2-1    ASP    .
            5    #2     @2      CYS    0.2
            6    #2     @2      LEU    0.8
            7    #2     @3      VAL    0.2
            8    #2     @3      SER    0.8

    In this case default is the name of the list, the list should have a name
    The chain_code and sequence_code should be for unassigned chains and residues, the residue code can include
    a sequence offset
    The labelled_residue flag indicates that the residue
    Probability can be used to give the probability of a particular residue type / . for the default probability
    which would be the 1 / <number-of-residues-in-group> [optional]
    these probabilities should be normalised by the program that uses them.

    note contradictory assignments should cause an error in the program reading the list
"""


@dataclass
class Unlabellinglevel:
    target_residues: AminoAcid
    level: float
    cross_residues: Dict[AminoAcid, float]


# these values are _estimated_ from the graphs in the supplementary material of Bellstedt, P. et al. Resonance
# assignment for a particularly challenging protein based on systematic unlabeling of amino acids to complement
# incomplete NMR data sets. J Biomol Nmr 57, 65 72 (2013).

#
# _RAW_SCRAMBLING_14N_GB1 = {
#     "Ala": (
#         0.0,
#         ("Val", 0.62,),
#         (ANY_RESIDUE, 1.15)
#     ),
#     "Arg": (
#         0.0,
#         (ANY_RESIDUE, 1.10),
#     ),
#     "Asn": (
#         0.0,
#         (ANY_RESIDUE, 0.9)
#     ),
#     "Asp": (
#         0.6,
#         (ANY_RESIDUE, 0.6)  # not included leads to uniform unlabelling
#     ),
#     "Cys": (
#         0.0,
#         (ANY_RESIDUE, 0.95)
#     ),
#
#     "Gln": (
#         0.02,
#         (ANY_RESIDUE, 0.3)
#     ),
#
#     "Glu": (
#         0.45,
#         (ANY_RESIDUE, 0.45)
#     ),
#
#     "Gly": (
#         0.0,
#         ("Trp", 0.5),
#         (ANY_RESIDUE, 0.95)
#     ),
#
#     "His": (
#         0.0,
#         (ANY_RESIDUE, 0.9)
#     ),
#     "Ile": (
#        0.25,
#        ("Leu", 0.4),
#        ("Val", 0.4),
#        (ANY_RESIDUE, 0.95)
#     ),
#     "Leu": (
#         0.19,
#         ('Ile', 0.2),
#         ('Val', 0.3),
#         (ANY_RESIDUE, 0.85)
#     ),
#     "Lys": (
#         0.02,
#         (ANY_RESIDUE, 1.10)
#     ),
#     "Met": (
#         0.0,
#         (ANY_RESIDUE, 1.10)
#     ),
#     "Phe": (
#         0.3,
#         ("Tyr", 0.35),
#         ("Ile", 0.78),
#         ('Leu', 0.8),
#         ('Val', 0.80),
#         (ANY_RESIDUE, 0.9)
#     ),
#     "Pro": (
#         0.0,
#         (ANY_RESIDUE, 0.90)
#     ),
#     "Ser": (
#         0.0,
#         (ANY_RESIDUE, 0.95)
#     ),
#     "Thr": (
#         0.0,
#         ("Gly", 0.0),
#         ("Trp", 0.5),
#         ("Lys", 0.90),
#         ("Tyr", 0.90),
#         (ANY_RESIDUE, 0.80),
#
#     ),
#     "Trp": (
#         0.2,
#         ("Phe", 0.3),
#         ("Tyr", 0.3),
#         (ANY_RESIDUE, 0.7)
#     ),
#     "Tyr": (
#         0.45,
#         ("Phe", 0.55),
#         (ANY_RESIDUE, 0.95)
#     ),
#     "Val": (
#         0.15,
#         ("Ile", 0.5),
#         ("Leu", 0.55),
#         ('Ala', 0.82),
#         (ANY_RESIDUE, 1.00)
#     ),
# }

_RAW_SCRAMBLING_14N_GB1 = {
    AminoAcid.ALA: (
        0.0,
        (
            AminoAcid.VAL,
            0.62,
        ),
        (AminoAcid.ANY, 1.15),
    ),
    AminoAcid.ARG: (
        0.0,
        (AminoAcid.ANY, 1.10),
    ),
    AminoAcid.ASN: (0.0, (AminoAcid.ANY, 0.9)),
    AminoAcid.ASP: (0.6, (AminoAcid.ANY, 0.6)),
    AminoAcid.CYS: (0.0, (AminoAcid.ANY, 0.95)),
    AminoAcid.GLN: (0.02, (AminoAcid.ANY, 0.3)),
    AminoAcid.GLU: (0.45, (AminoAcid.ANY, 0.45)),
    AminoAcid.GLY: (0.0, (AminoAcid.TRP, 0.5), (AminoAcid.ANY, 0.95)),
    AminoAcid.HIS: (0.0, (AminoAcid.ANY, 0.9)),
    AminoAcid.ILE: (
        0.25,
        (AminoAcid.LEU, 0.4),
        (AminoAcid.VAL, 0.4),
        (AminoAcid.ANY, 0.95),
    ),
    AminoAcid.LEU: (
        0.19,
        (AminoAcid.ILE, 0.2),
        (AminoAcid.VAL, 0.3),
        (AminoAcid.ANY, 0.85),
    ),
    AminoAcid.LYS: (0.02, (AminoAcid.ANY, 1.10)),
    AminoAcid.MET: (0.0, (AminoAcid.ANY, 1.10)),
    AminoAcid.PHE: (
        0.3,
        (AminoAcid.TYR, 0.35),
        (AminoAcid.ILE, 0.78),
        (AminoAcid.LEU, 0.8),
        (AminoAcid.VAL, 0.80),
        (AminoAcid.ANY, 0.9),
    ),
    AminoAcid.PRO: (0.0, (AminoAcid.ANY, 0.90)),
    AminoAcid.SER: (0.0, (AminoAcid.ANY, 0.95)),
    AminoAcid.THR: (
        0.0,
        (AminoAcid.GLY, 0.0),
        (AminoAcid.TRP, 0.5),
        (AminoAcid.LYS, 0.90),
        (AminoAcid.TYR, 0.90),
        (AminoAcid.ANY, 0.80),
    ),
    AminoAcid.TRP: (
        0.2,
        (AminoAcid.PHE, 0.3),
        (AminoAcid.TYR, 0.3),
        (AminoAcid.ANY, 0.7),
    ),
    AminoAcid.TYR: (0.45, (AminoAcid.PHE, 0.55), (AminoAcid.ANY, 0.95)),
    AminoAcid.VAL: (
        0.15,
        (AminoAcid.ILE, 0.5),
        (AminoAcid.LEU, 0.55),
        (AminoAcid.ALA, 0.82),
        (AminoAcid.ANY, 1.00),
    ),
}

SCRAMBLING_14N_GB1 = {
    residue_name: Unlabellinglevel(residue_name, level, dict(cross_residues))
    for residue_name, (level, *cross_residues) in _RAW_SCRAMBLING_14N_GB1.items()
}


# _RAW_SCRAMBLING_14N_APRAYAXIN = {
#     "Ala": (
#         0.0,
#         ("Val", 0.50,),
#         (ANY_RESIDUE, 0.85)
#     ),
#     "Arg": (
#         0.0,
#         (ANY_RESIDUE, 0.95),
#     ),
#     "Asn": (
#         0.0,
#         (ANY_RESIDUE, 0.88)
#     ),
#     "Asp": (
#         0.45,
#         (ANY_RESIDUE, 0.45)  # not included leads to uniform unlabelling
#     ),
#     "Cys": (
#         0.0,
#         ("Ala", 0.3),
#         ("Arg", 0.7),
#         ("Lys", 0.55),
#         ("Thr", 0.75),
#         ("Trp", 0.85),
#         ("Val", 0.85),
#         (ANY_RESIDUE, 0.95)
#     ),
#     "Gln": (
#         0.05,
#         (ANY_RESIDUE, 0.25)
#     ),
#     "Glu": (
#         0.45,
#         (ANY_RESIDUE, 0.45)
#     ),
#     "Gly": (
#         0.0,
#         ("Ala", 0.85),
#         ("Cys", 0.0),
#         ("Ser", 0.0),
#         ("Trp", 0.60),
#         (ANY_RESIDUE, 0.95),
#     ),
#     "His": (
#         0.0,
#         (ANY_RESIDUE, 0.85)
#     ),
#     "Ile": (
#         0.20,
#         ("Ala", 0.65),
#         ("Leu", 0.325),
#         ("Val", 0.35),
#         (ANY_RESIDUE, 0.80)
#     ),
#     'Leu': (
#         0.20,
#         ('Ala', 0.65),
#         ('Ile', 0.20),
#         ('Val', 0.35),
#         (ANY_RESIDUE, 0.80)
#     ),
#     "Lys": (
#         0.02,
#         (ANY_RESIDUE, 0.90)
#     ),
#     "Met": (
#         0.3,
#         ("Phe", 0.80),
#         ("Tyr", 0.80),
#         (ANY_RESIDUE, 0.90)
#     ),
#     "Pro": (
#         0.0,
#         (ANY_RESIDUE, 0.70)
#     ),
#     "Ser": (
#         0.55,
#         (ANY_RESIDUE, 0.55)
#     ),
#     "Thr": (
#         0.0,
#         ("Cys", 0.25),
#         ("Gly", 0.0),
#         ("Ser", 0.05),
#         ("Trp", 0.70),
#         (ANY_RESIDUE, 0.90)
#     ),
#     "Trp": (
#         0.45,
#         ("Phe", 0.6),
#         ("Tyr", 0.3),
#         (ANY_RESIDUE, 0.80)
#     ),
#     "Tyr": (
#         0.50,
#         ("Phe", 0.65),
#         ("Trp", 0.7),
#         (ANY_RESIDUE, 0.80)
#     ),
#     "Val": (
#         0.15,
#         ("Ala", 0.65),
#         ("Ile", 0.45),
#         ("Leu", 0.50),
#         (ANY_RESIDUE, 0.85)
#     ),
# }

_RAW_SCRAMBLING_14N_APRAYAXIN = {
    AminoAcid.ALA: (
        0.0,
        (
            AminoAcid.VAL,
            0.50,
        ),
        (AminoAcid.ANY, 0.85),
    ),
    AminoAcid.ARG: (
        0.0,
        (AminoAcid.ANY, 0.95),
    ),
    AminoAcid.ASN: (0.0, (AminoAcid.ANY, 0.88)),
    AminoAcid.ASP: (
        0.45,
        (AminoAcid.ANY, 0.45),  # not included leads to uniform unlabelling
    ),
    AminoAcid.CYS: (
        0.0,
        (AminoAcid.ALA, 0.3),
        (AminoAcid.ARG, 0.7),
        (AminoAcid.LYS, 0.55),
        (AminoAcid.THR, 0.75),
        (AminoAcid.TRP, 0.85),
        (AminoAcid.VAL, 0.85),
        (AminoAcid.ANY, 0.95),
    ),
    AminoAcid.GLN: (0.05, (AminoAcid.ANY, 0.25)),
    AminoAcid.GLU: (0.45, (AminoAcid.ANY, 0.45)),
    AminoAcid.GLY: (
        0.0,
        (AminoAcid.ALA, 0.85),
        (AminoAcid.CYS, 0.0),
        (AminoAcid.SER, 0.0),
        (AminoAcid.TRP, 0.60),
        (AminoAcid.ANY, 0.95),
    ),
    AminoAcid.HIS: (0.0, (AminoAcid.ANY, 0.85)),
    AminoAcid.ILE: (
        0.20,
        (AminoAcid.ALA, 0.65),
        (AminoAcid.LEU, 0.325),
        (AminoAcid.VAL, 0.35),
        (AminoAcid.ANY, 0.80),
    ),
    AminoAcid.LEU: (
        0.20,
        (AminoAcid.ALA, 0.65),
        (AminoAcid.ILE, 0.20),
        (AminoAcid.VAL, 0.35),
        (AminoAcid.ANY, 0.80),
    ),
    AminoAcid.LYS: (0.02, (AminoAcid.ANY, 0.90)),
    AminoAcid.MET: (
        0.3,
        (AminoAcid.PHE, 0.80),
        (AminoAcid.TYR, 0.80),
        (AminoAcid.ANY, 0.90),
    ),
    AminoAcid.PRO: (0.0, (AminoAcid.ANY, 0.70)),
    AminoAcid.SER: (0.55, (AminoAcid.ANY, 0.55)),
    AminoAcid.THR: (
        0.0,
        (AminoAcid.CYS, 0.25),
        (AminoAcid.GLY, 0.0),
        (AminoAcid.SER, 0.05),
        (AminoAcid.TRP, 0.70),
        (AminoAcid.ANY, 0.90),
    ),
    AminoAcid.TRP: (
        0.45,
        (AminoAcid.PHE, 0.6),
        (AminoAcid.TYR, 0.3),
        (AminoAcid.ANY, 0.80),
    ),
    AminoAcid.TYR: (
        0.50,
        (AminoAcid.PHE, 0.65),
        (AminoAcid.TRP, 0.7),
        (AminoAcid.ANY, 0.80),
    ),
    AminoAcid.VAL: (
        0.15,
        (AminoAcid.ALA, 0.65),
        (AminoAcid.ILE, 0.45),
        (AminoAcid.LEU, 0.50),
        (AminoAcid.ANY, 0.85),
    ),
}

SCRAMBLING_14N_APRATAXIN = {
    residue_name: Unlabellinglevel(residue_name, level, dict(cross_residues))
    for residue_name, (level, *cross_residues) in _RAW_SCRAMBLING_14N_APRAYAXIN.items()
}


@simulate_app.command()
def unlabelling(
    context: typer.Context,
    input: Path = typer.Option(
        Path("-"),
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="where to read NEF data from either a file or stdin '-'",
    ),
    frame_selectors: List[str] = typer.Option(
        None, "--frames", metavar=None, help=FRAME_SELECTORS_HELP
    ),
    exact: bool = typer.Option(False, "--exact", help="match the frame names exactly"),
    name_template: str = typer.Option(
        UNLABELLED_SPECTRUM_NAME_TEMPLATE, help=NAME_TEMPLATE_HELP
    ),
    list_sets: bool = typer.Option(
        False, "--list", help="list the available amino acid sets"
    ),
    residue_types: bool = typer.Option(
        False, "--residue-types", help=RESIDUE_TYPES_HELP
    ),
    display_residue_types_format: bool = typer.Option(
        False, "--residue-type-format", help=DISPLAY_RESIDUE_TYPES_FORMAT_HELP
    ),
    unlabelled_amino_acids_and_sets: List[str] = typer.Argument(
        None, help=UNLABELLED_AMINO_ACIDS_AND_SETS_HELP
    ),
):
    """-  make a set of synthetic peaks for a 14N/15N unlabelling experiment NMR using input shifts or peaks
    data from Bellstedt et al. J. Biomol. Nmr 57, 65-72 (2013) doi:10.1007/s10858-013-9768-0
    """

    _list_labeling_sets_and_exit_if_list_true(list_sets)

    _display_residue_types_format_and_exit_if_format_true(display_residue_types_format)

    unlabelled_amino_acids_and_sets = parse_comma_separated_options(
        unlabelled_amino_acids_and_sets
    )

    unlabelled_amino_acids_and_sets = _expand_labelling_sets(
        unlabelled_amino_acids_and_sets
    )

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    all_unknown_amino_acids = {
        amino_acid
        for amino_acid in unlabelled_amino_acids_and_sets
        if amino_acid.upper() not in ALL_AMINO_ACIDS
    }

    _exit_if_residues_are_unknown(all_unknown_amino_acids)

    if not unlabelled_amino_acids_and_sets:
        unlabelled_amino_acids_and_sets = NO_OBV_CLEAN_SET

    if not frame_selectors:
        # TODO exact is not used here!
        if residue_types:
            selected_frames = select_frames(
                entry, "nef_chemical_shift_list_default", SelectionType.NAME
            )
        else:
            selected_frames = select_frames(
                entry, "nef_nmr_spectrum", SelectionType.CATEGORY
            )

    else:
        selected_frames = select_frames_by_name(entry, frame_selectors, exact=exact)

        msg = f"""
            using the selectors {', '.join(frame_selectors)} and with residue_types {residue_types}
            I couldn't find any frames to use in the entry {entry.entry_id}
        """
        exit_error(msg)

    entry = pipe(
        entry,
        selected_frames,
        unlabelled_amino_acids_and_sets,
        name_template,
        output_residue_typing=residue_types,
    )

    print(entry)


def pipe(
    entry: Entry,
    selected_frames: List[Saveframe],
    unlabelled_residues: List[str],
    name_template_string: str,
    output_residue_typing: bool = False,
) -> Entry:

    # this could be almost anyting that has a sequence code and a residue name
    shift_frames = [
        frame
        for frame in selected_frames
        if frame.category in {"nef_chemical_shift_list", "nef_molecular_system"}
    ]
    peak_frames = [
        frame for frame in selected_frames if frame.category == "nef_nmr_spectrum"
    ]

    if not shift_frames and not peak_frames:
        exit_error("working with peak frames not implemented yet..., bug gary!")

    frames = []

    if output_residue_typing:
        if peak_frames:
            msg = """
                I can't output residue types from peak frames currently (bug gary!)
            """
            exit_error(msg)

        frames.append(_make_residue_typing_table(shift_frames, unlabelled_residues))

    else:
        peak_lists = {}
        spectrum_info = EXPERIMENT_INFO[SPECTRUM_TYPE_NHSQC]
        if shift_frames:
            # TODO different naming for peak_lists and shifts frame functions
            shifts = nef_frames_to_shifts(shift_frames)

            shifts = [shift for shift in shifts if shift.atom.atom_name in ["H", "N"]]

            shifts = _average_equivalent_shifts(shifts)

            peak_lists["shifts"] = _make_peak_list(shifts, spectrum_info)

        elif peak_frames:
            for peak_frame in peak_frames:
                peak_lists[peak_frame.name] = frame_to_peaks(peak_frame)
        else:
            msg = """
                No peak_lists or frames were selected to base the ¹H-¹⁵N HSQC to unlabelled spectrum on
            """
            exit_error(msg)

        # todo check there are some 1H 15N pairs

        if peak_lists:
            for peak_list_name, peak_list in peak_lists.items():
                for target_residue in unlabelled_residues:
                    # for template expansion
                    residues = target_residue  # noqa: F841
                    spectrum = peak_list_name  # noqa: F841
                    frame_name = f(name_template_string)

                    frames.append(
                        _make_spectrum_frame(
                            peak_list, target_residue, spectrum_info, frame_name
                        )
                    )
        else:

            cause = "shifts" if shift_frames else "peak_lists"

            exit_error(
                f"couldn't build a ¹H-¹⁵N HSQC to unlabel because i couldn't find any {cause}"
            )

    for frame in frames:
        entry.add_saveframe(frame)

    return entry


def _make_residue_typing_table(selected_frames, unlabelled_residues):

    frame = create_nef_save_frame(f"{NEF_PIPELINES_NAMESPACE}_residue_types", "default")

    loop = Loop.from_scratch(f"{NEF_PIPELINES_NAMESPACE}_residue_type")

    for tag in ["index", "chain_code", "sequence_code", "residue_type", "probability"]:
        loop.add_tag(tag)

    frame.add_loop(loop)
    sequence = sequences_from_frames(selected_frames)
    residues_to_list = []

    for residue in sequence:
        if residue.residue_name in unlabelled_residues:
            residues_to_list.append(residue)

    for index, residue in enumerate(residues_to_list, start=1):
        row = {
            "index": index,
            "chain_code": residue.chain_code,
            "sequence_code": residue.sequence_code,
            "residue_type": residue.residue_name,
            "probability": UNUSED,
        }
        loop.add_data(
            [
                row,
            ]
        )

    return frame


def _make_spectrum_frame(
    peaks,
    target_residue,
    spectrum_info,
    frame_name,
    spectrometer_frequency=FAKE_SPECTROMETER_FREQUENCY_600,
):
    if target_residue == AminoAcid.REF:
        peaks = deepcopy(peaks)
    else:
        peaks = _unlabel_peaks(
            deepcopy(peaks),
            target_residue,
            [SCRAMBLING_14N_APRATAXIN, SCRAMBLING_14N_GB1],
        )

    dimensions = [{"axis_code": dimension} for dimension in spectrum_info.dimensions]
    spectrum_classification = SPECTRUM_TYPE_TO_CLASSIFICATION[ExperimentType.N_HSQC]
    extra_tags = {
        EXPERIMENT_CLASSIFICATION: spectrum_classification,
        EXPERIMENT_TYPE: EXPERIMENT_CLASSIFICATION_TO_SYNONYM[spectrum_classification],
    }
    return peaks_to_frame(
        peaks, dimensions, spectrometer_frequency, frame_name, extra_tags=extra_tags
    )


def _average_equivalent_shifts(shifts):

    shifts_by_atom = {}

    for shift in shifts:
        atom = shift.atom

        shifts_by_atom.setdefault(atom, []).append(shift)

    for atom, shifts in shifts_by_atom.items():
        if len(shifts) > 1:
            shifts = [ufloat(shift.value, shift.value_uncertainty) for shift in shifts]
            mean_shift = sum(shift) / len(shifts)

            new_shift = replace(
                shifts[0],
                value=mean_shift.nominal_value,
                uncertainty=mean_shift.std_dev,
            )

            shifts_by_atom[atom] = [
                new_shift,
            ]

    return [shift[0] for shift in shifts_by_atom.values()]


def _unlabel_peaks(peaks, unlabeled_residue, unlabelling_patterns):

    pattern_for_residue = None
    for pattern in unlabelling_patterns:
        if unlabeled_residue in pattern:
            pattern_for_residue = pattern[unlabeled_residue]

    cross_residues = pattern_for_residue.cross_residues
    general_modulation = pattern_for_residue.cross_residues[AminoAcid.ANY]

    modulated_peaks = []
    for peak in peaks:
        residue_names = set([shift.atom.residue.residue_name for shift in peak.shifts])

        if len(residue_names) != 1:
            exit_error(
                "INTERNAL ERROR: unlabelling is only supported for peaks with a single residue type"
            )

        residue = residue_names.pop()

        height = peak.height
        volume = peak.volume
        if residue == unlabeled_residue:
            height *= pattern_for_residue.level
            volume *= pattern_for_residue.level
        else:
            modulation = (
                cross_residues[residue]
                if residue in cross_residues
                else general_modulation
            )
            height *= modulation
            volume *= modulation

        modulated_peaks.append(replace(peak, height=height, volume=volume))

    return modulated_peaks


def _exit_if_residues_are_unknown(all_unknown_amino_acids):
    if all_unknown_amino_acids:
        msg = f"""\
            I don't have unlabelling information for the following residues:
            {', '.join(all_unknown_amino_acids)}
        """
        exit_error(msg)


def _display_residue_types_format_and_exit_if_format_true(format: bool):
    if format:
        print(RESIDUE_TYPES_FORMAT, file=sys.stderr)
        print()
        print("exiting...")
        sys.exit(0)


def _expand_labelling_sets(unlabelled_amino_acids_and_sets):
    expanded_unlabelled_amino_acids_and_sets = set()
    for residue_or_set in unlabelled_amino_acids_and_sets:
        if residue_or_set in LABELING_SETS:
            if residue_or_set == "REMOVE_OBVIOUS":
                expanded_unlabelled_amino_acids_and_sets -= OBVIOUS_SHIFT_SET
            else:
                expanded_unlabelled_amino_acids_and_sets.update(
                    LABELING_SETS[residue_or_set]
                )
        else:
            expanded_unlabelled_amino_acids_and_sets.add(residue_or_set)
    return expanded_unlabelled_amino_acids_and_sets


def _list_labeling_sets_and_exit_if_list_true(list: bool):
    if list:
        table = []
        for name, description in LABELING_SET_HELP.items():
            table.append([name, description])
            amino_acids = [value.name for value in LABELING_SETS[name]]
            table.append(["", ", ".join(amino_acids)])
            table.append(["", ""])
        print()
        print(tabulate(table, headers=["Name", "Description"]))
        print("\nexiting...")
        sys.exit(0)



#!/usr/bin/env python3

import logging
import sys
from importlib import import_module
from textwrap import dedent
from traceback import format_exc, print_exc

verbose_mode = False

try:
    import future  # noqa: F401
except Exception:
    pass

MINIMUM_VERSION = (3, 7)
if not sys.version_info >= MINIMUM_VERSION:
    print(
        "Error; minimum version required id %i.%i [and python 2 is not supported]"
        % MINIMUM_VERSION,
        file=sys.stderr,
    )
    print("exiting...", file=sys.stderr)

# for pynmrstar to avoid WARNING:root:Loop with no data on line: xx
logging.getLogger("pynmrstar").setLevel(logging.ERROR)

EXIT_ERROR = 1


def do_exit_error(msg, trace_back=True, exit_code=EXIT_ERROR):
    msg = dedent(msg)
    if trace_back:
        print_exc()
    print(msg, file=sys.stderr)
    print("exiting...", file=sys.stderr)
    sys.exit(exit_code)


def create_nef_app():
    import typer

    from nef_pipelines import nef_app

    nef_app.app = typer.Typer(no_args_is_help=True)
    app = nef_app.app  # noqa: F841
    return nef_app


def main():
    global verbose_mode
    try:
        import typer
        from click import ClickException

    except Exception as e:

        msg = """\

             Initializaion error: one of the core libraries [friendly_tracback/typer] if missing from you environment
             please make sure they are installed an try again
             exiting..."""
        do_exit_error(msg, e)

    try:
        nef_app = create_nef_app()

    except Exception as e:
        msg = """\

                 Initialisation error: failed to start the typer app, message the developer
                 exiting..."""
        do_exit_error(msg, e)

    warnings = []
    try:
        # import components which will self register, this could and will be automated
        modules = [
            "nef_pipelines.tools.help",
            "nef_pipelines.tools.chains",
            "nef_pipelines.tools.entry",
            "nef_pipelines.tools.fit",
            "nef_pipelines.tools.frames",
            "nef_pipelines.tools.globals",
            "nef_pipelines.tools.header",
            "nef_pipelines.tools.loops",
            "nef_pipelines.tools.peaks",
            "nef_pipelines.tools.save",
            "nef_pipelines.tools.series",
            "nef_pipelines.tools.shifts",
            "nef_pipelines.tools.simulate",
            "nef_pipelines.tools.sink",
            "nef_pipelines.tools.stream",
            "nef_pipelines.tools.test",
            "nef_pipelines.transcoders.csv",
            "nef_pipelines.transcoders.deep",
            "nef_pipelines.transcoders.echidna",
            "nef_pipelines.transcoders.fasta",
            "nef_pipelines.transcoders.mars",
            "nef_pipelines.transcoders.modelfree",
            "nef_pipelines.transcoders.nmrpipe",
            "nef_pipelines.transcoders.nmrview",
            "nef_pipelines.transcoders.pales",
            "nef_pipelines.transcoders.rcsb",
            "nef_pipelines.transcoders.rpf",
            "nef_pipelines.transcoders.shifty",
            "nef_pipelines.transcoders.shiftx2",
            "nef_pipelines.transcoders.sparky",
            "nef_pipelines.transcoders.nmrstar",
            "nef_pipelines.transcoders.talos",
            "nef_pipelines.transcoders.xcamshift",
            "nef_pipelines.transcoders.xeasy",
            "nef_pipelines.transcoders.xplor",
        ]
        for module_name in modules:
            try:
                import_module(module_name)
            except Exception:
                msg = f"plugin {module_name}\n{format_exc()}"

                warnings.append((module_name, msg))

    except Exception as e:
        msg = """\

             Initialisation error: failed to load a plugin, remove the plugin or contact the developer
             """

        do_exit_error(msg, e)

    try:

        nef_app.app

        command = typer.main.get_command(nef_app.app)

        command(standalone_mode=False)

        _report_warnings(warnings)

    except ClickException as e:
        e.show()
        do_exit_error(
            f"inputs: {' '.join(sys.argv[1:])}", trace_back=False, exit_code=e.exit_code
        )

    except Exception as e:

        msg = f"""\

              Runtime error: failed to process the data using the plugin and commands, check you inputs or report a bug
              inputs: {' '.join(sys.argv[1:])}
              message: {e}
              """

        do_exit_error(msg)


def _report_warnings(warnings):
    if warnings:

        bad_modules = []
        for module_name, warning in warnings:
            bad_modules.append(module_name)

            lines = warning.split("\n")
            max_line_length = max([len(line) for line in lines])
            msg = f"error in {module_name}"
            line_length_m_module_name = max_line_length - len(msg)
            stars_2_m1 = (line_length_m_module_name // 2) - 2

            header = f"{'*' * stars_2_m1} {msg} {'*' * stars_2_m1}"
            discrepancy = max_line_length - len(header)
            header = f'{header}{"*" * discrepancy}'

            print(file=sys.stderr)
            print(header)
            print(file=sys.stderr)
            print(warning, file=sys.stderr)
            print("*" * max_line_length)

        NEW_LINE = "\n    "
        msg = f"""
                WARNING: the following plugins failed to load:

                    {NEW_LINE.join(bad_modules)}

                the remaining plugins will work but NEF-Pipelines is working  with reduced capabilities. The modules
                that failed to load and the causes of the problem are listed above please report this to the authors
                at github: https://github.com/varioustoxins/NEF-Pipelines/issues
            """
        msg = dedent(msg)
        print(msg, file=sys.stderr)

    # # to get rid of messages about broken pipes when SIGPIPE is recieved
    # Exception ignored in: <_io.TextIOWrapper name='<stdout>' mode='w' encoding='utf-8'>
    # BrokenPipeError: [Errno 32] Broken pipe
    # https://stackoverflow.com/questions/26692284/how-to-prevent-brokenpipeerror-when-doing-a-flush-in-python
    # note: https://docs.python.org/3/library/signal.html#note-on-sigpipe
    #      doesn't appear to work for us, not quite sure why
    # note: I added a flush of stderr to remove any dangling output
    sys.stderr.flush()
    sys.stderr.close()


if __name__ == "__main__":
    main()



from itertools import zip_longest
from pathlib import Path
from typing import Dict, List

import typer

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    loop_row_namespace_iter,
    molecular_system_from_entry,
    molecular_system_from_entry_or_exit,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (
    chains_from_frames,
    get_chain_starts,
    get_residue_name_from_lookup,
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
    sequences_from_frames,
)
from nef_pipelines.lib.structures import AtomLabel, SequenceResidue, ShiftData
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    is_float,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.fasta.exporters.sequence import STDOUT
from nef_pipelines.transcoders.nmrview import export_app

NEF_CHEMICAL_SHIFT = "nef_chemical_shift"

NEF_CHEMICAL_SHIFT_LIST = "nef_chemical_shift_list"

app = typer.Typer()


@export_app.command()
def shifts(
    file_name_template: str = typer.Option(
        "%s.out",
        "-t",
        "--template",
        help="the template for the filename to export to %s will get replaced by the name of the chain or a filename if"
        "set with the sequence_text-names option",
        metavar="<sequence-sequence_text.seq>",
    ),
    input: Path = typer.Option(
        STDIN, "-i", "--in", help="sequence_text to read input from [- is stdin]"
    ),
    output_to_stdout: bool = typer.Option(
        False, "-o", "--out", help="write the files to stdout for debugging"
    ),
    file_names: List[str] = typer.Option(
        None,
        "--file-names",
        help="alternative filenames to export to, can be a comma "
        "separated list of filenames. Repeated calls add "
        "more filenames.",
    ),
    split_lists: bool = typer.Option(
        False, "-c", "--combine", help="merge all shift lists into one"
    ),
    chain_selectors: List[str] = typer.Argument(
        None,
        help="the names of the chains to export, can be a comma separated"
        " list of chains. Repeated calls add more chains. The default"
        " is to write all chains",
        metavar="<chains>",
    ),
):

    chain_selectors = parse_comma_separated_options(chain_selectors)

    file_names = parse_comma_separated_options(file_names)

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    molecular_system_frame = molecular_system_from_entry_or_exit(entry)

    chain_codes = chains_from_frames(molecular_system_frame)

    _exit_if_chain_selector_not_in_chain_codes(chain_codes, chain_selectors)

    chain_selectors = chain_selectors or chain_codes

    _exit_if_more_file_names_than_chains(chain_selectors, file_names)

    pipe(
        entry,
        output_to_stdout,
        file_name_template,
        file_names,
        split_lists,
        chain_selectors,
    )


def pipe(
    entry,
    output_to_stdout,
    file_name_template,
    file_names,
    split_lists,
    chain_selectors,
):

    sequence = sequence_from_entry_or_exit(entry)

    residue_name_lookup = sequence_to_residue_name_lookup(sequence)

    shifts = read_all_defined_shifts_from_entry(entry, residue_name_lookup)

    print(shifts)
    # chains_to_filenames = _build_chains_to_filenames(
    #     chain_selectors, file_names, file_name_template
    # )
    #
    # chain_sequences = _nmrview_sequences_from_entry(entry, chain_selectors)
    #
    # chains_to_output_file_names = _chains_to_output_filenames(
    #     chains_to_filenames, output_to_stdout
    # )
    #
    # file_banners, final_banner = _make_file_name_banners(chains_to_filenames)
    #
    # for chain_code, sequence_text in chain_sequences.items():
    #
    #     output_file_name = chains_to_output_file_names[chain_code]
    #
    #     with smart_open(output_file_name) as file_pointer:
    #
    #         if output_to_stdout:
    #             print(file_banners[chain_code], file=file_pointer)
    #
    #         print(sequence_text, file=file_pointer)
    # if output_to_stdout:
    #     print(final_banner)
    # elif not stdout.isatty():
    #     print(entry)


def read_all_defined_shifts_from_entry(entry, residue_name_lookup):
    save_frames = entry.get_saveframes_by_category(NEF_CHEMICAL_SHIFT_LIST)
    result = []
    for save_frame in save_frames:

        loop = save_frame.get_loop(NEF_CHEMICAL_SHIFT)
        for row in loop_row_namespace_iter(loop):
            chain_code = row.chain_code
            sequence_code = row.sequence_code

            residue_name = get_residue_name_from_lookup(
                chain_code, sequence_code, residue_name_lookup
            )
            residue = SequenceResidue(chain_code, sequence_code, residue_name)
            atom_name = row.atom_name
            atom = AtomLabel(residue, atom_name)

            if _is_complete_atom(atom):
                value = row.value
                if value != UNUSED and is_float(value):
                    uncertainty = row.value_uncertainty
                    shift = ShiftData(atom, value, uncertainty)
                    result.append(shift)
    return result


def _is_complete_atom(atom: AtomLabel) -> bool:
    ok = True

    residue = atom.residue
    if residue.chain_code == UNUSED:
        ok = False
    if residue.sequence_code == UNUSED:
        ok = False
    if residue.residue_name == UNUSED:
        ok = False
    if atom.atom_name == UNUSED:
        ok = False

    return ok


def _make_file_name_banners(chains_to_filenames):
    same_width_names = _make_chain_code_to_names_same_width_names(chains_to_filenames)

    file_banners = {
        chain_code: f"------------- {same_width_names[chain_code]} -------------"
        for chain_code, file_name in chains_to_filenames.items()
    }
    banner_lengths = [len(banner) for banner in file_banners.values()]
    max_banner_width = max(banner_lengths)

    final_banner = "-" * max_banner_width

    return file_banners, final_banner


def _chains_to_output_filenames(chains_to_filenames, output_to_stdout):
    return {
        chain: STDOUT if output_to_stdout else file_name
        for chain, file_name in chains_to_filenames.items()
    }


def _exit_if_chain_selector_not_in_chain_codes(chain_codes, chain_selectors):
    for chain_selector in chain_selectors:
        if chain_selector not in chain_codes:
            msg = f"""\
                the chain code {chain_selector} is not in the molecular systems chain codes
                chain selectors:  {', '.join(chain_selectors)}
                chain codes: {', '.join(chain_codes)}
            """
        exit_error(msg)


def _exit_if_more_file_names_than_chains(chain_selectors, file_names):
    if len(file_names) > len(chain_selectors):
        msg = f"""\
            there are more file names than chains !

            chains = {','.join(chain_selectors)} [{len(chain_selectors)}]
            file names: {''.join(file_names)} [{len(file_names)}]
        """

        exit_error(msg)


def _nef_to_nmrview_sequences(
    residues: List[SequenceResidue], chain_codes: List[str]
) -> Dict[str, List[str]]:

    chain_starts = get_chain_starts(residues)

    residues_by_chain = {
        chain_code: [
            residue for residue in residues if residue.chain_code == chain_code
        ]
        for chain_code in chain_codes
    }
    residue_sequences = {
        chain_code: [residue.residue_name for residue in residues_by_chain[chain_code]]
        for chain_code in chain_codes
    }
    result = {}
    for chain_code, residue_sequence in residue_sequences.items():
        chain_start = chain_starts[chain_code]

        sequence_lines = []
        for i, residue in enumerate(residue_sequence):

            residue = residue.lower()

            if i == 0:
                if chain_start != 1:
                    line = f"{residue} {chain_start}"
            else:
                line = residue

            sequence_lines.append(line)

        result[chain_code] = "\n".join(sequence_lines)

    return result


def _nmrview_sequences_from_entry(entry, chain_codes):

    molecular_system_frame = molecular_system_from_entry(entry)

    residues = sequences_from_frames(molecular_system_frame)

    return _nef_to_nmrview_sequences(residues, chain_codes)


def _make_chain_code_to_names_same_width_names(names: Dict[str, str]) -> Dict[str, str]:
    max_length = -1
    for name in names.values():
        name_len = len(name)
        max_length = max(name_len, max_length)

    return {chain_code: name.center(max_length) for chain_code, name in names.items()}


def _open_for_writing_or_exit(file_name):
    try:
        result = open(file_name, "w")
    except Exception as e:
        msg = f"failed to open {file_name} for writing because {e}"

        exit_error(msg, e)

    return result


def _build_chains_to_filenames(
    chain_selectors: str, file_names: str, file_name_template: str
) -> Dict[str, str]:

    return {
        chain_selector: file_name or file_name_template % chain_selector
        for chain_selector, file_name in zip_longest(chain_selectors, file_names)
    }



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.sparky.importers.sequence import sequence

app = typer.Typer()
app.command()(sequence)

EXPECTED_DNA = """\
  loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1    A   2    DT   start    .   .
     2    A   3    DG   middle   .   .
     3    A   4    DC   middle   .   .
     4    A   5    DA   middle   .   .
     5    A   6    DT   middle   .   .
     6    A   7    DG   middle   .   .
     7    A   8    DC   middle   .   .
     8    A   9    DA   middle   .   .
     9    A   10   DT   middle   .   .
     10   A   11   DG   middle   .   .
     11   A   12   DC   middle   .   .
     12   A   13   DA   middle   .   .
     13   A   14   DT   middle   .   .
     14   A   15   DG   middle   .   .
     15   A   16   DG   middle   .   .
     16   A   17   DT   middle   .   .
     17   A   18   DG   middle   .   .
     18   A   19   DC   middle   .   .
     19   A   20   DA   middle   .   .
     20   A   21   DT   end      .   .

   stop_
"""


def test_basic_dna():

    sequence_path = path_in_test_data(__file__, "sparky_basic_sequence_dna.txt")

    result = run_and_report(app, ["--molecule-type", "dna", sequence_path])

    assert_lines_match(
        EXPECTED_DNA,
        isolate_loop(result.stdout, "nef_molecular_system", "nef_sequence"),
    )


EXPECTED_PROTEIN = """
   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide
     1   A   2   ALA   start    .   .
     2   A   3   CYS   middle   .   .
     3   A   4   ASP   middle   .   .
     4   A   5   GLU   middle   .   .
     5   A   6   PHE   middle   .   .
     6   A   7   GLY   middle   .   .
     7   A   8   HIS   end      .   .
   stop_
"""


def test_basic_protein():

    sequence_path = path_in_test_data(__file__, "sparky_basic_sequence_protein.txt")

    result = run_and_report(app, [sequence_path])

    print(result.stdout)

    assert_lines_match(
        EXPECTED_PROTEIN,
        isolate_loop(result.stdout, "nef_molecular_system", "nef_sequence"),
    )



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.fit.t1noe import t1noe

runner = CliRunner()
app = typer.Typer()
app.command()(t1noe)


EXPECTED_R1_DATA_SINGLE_R1 = """
   loop_
      _nefpls_relaxation.index
      _nefpls_relaxation.data_id
      _nefpls_relaxation.data_combination_id
      _nefpls_relaxation.value
      _nefpls_relaxation.value_error
      _nefpls_relaxation.chain_code_1
      _nefpls_relaxation.sequence_code_1
      _nefpls_relaxation.residue_name_1
      _nefpls_relaxation.atom_name_1
      _nefpls_relaxation.chain_code_2
      _nefpls_relaxation.sequence_code_2
      _nefpls_relaxation.residue_name_2
      _nefpls_relaxation.atom_name_2

     1   1   .   1.300000   .   B   145   THR   H   B   145   THR   N

   stop_
"""

EXPECTED_R1_DATA_SINGLE_NOE = """
   loop_
      _nefpls_relaxation.index
      _nefpls_relaxation.data_id
      _nefpls_relaxation.data_combination_id
      _nefpls_relaxation.value
      _nefpls_relaxation.value_error
      _nefpls_relaxation.chain_code_1
      _nefpls_relaxation.sequence_code_1
      _nefpls_relaxation.residue_name_1
      _nefpls_relaxation.atom_name_1
      _nefpls_relaxation.chain_code_2
      _nefpls_relaxation.sequence_code_2
      _nefpls_relaxation.residue_name_2
      _nefpls_relaxation.atom_name_2

     1   1   .   1.000000   .   B   145   THR   H   B   145   THR   N

   stop_
"""


EXPECTED_R1_DATA_SINGLE_R1_MC = """
   loop_
      _nefpls_relaxation.index
      _nefpls_relaxation.data_id
      _nefpls_relaxation.data_combination_id
      _nefpls_relaxation.value
      _nefpls_relaxation.value_error
      _nefpls_relaxation.chain_code_1
      _nefpls_relaxation.sequence_code_1
      _nefpls_relaxation.residue_name_1
      _nefpls_relaxation.atom_name_1
      _nefpls_relaxation.chain_code_2
      _nefpls_relaxation.sequence_code_2
      _nefpls_relaxation.residue_name_2
      _nefpls_relaxation.atom_name_2

     1   1   .   1.300000   0.108797   B   145   THR   H   B   145   THR   N

   stop_
"""

EXPECTED_R1_DATA_SINGLE_NOE_MC = """
   loop_
      _nefpls_relaxation.index
      _nefpls_relaxation.data_id
      _nefpls_relaxation.data_combination_id
      _nefpls_relaxation.value
      _nefpls_relaxation.value_error
      _nefpls_relaxation.chain_code_1
      _nefpls_relaxation.sequence_code_1
      _nefpls_relaxation.residue_name_1
      _nefpls_relaxation.atom_name_1
      _nefpls_relaxation.chain_code_2
      _nefpls_relaxation.sequence_code_2
      _nefpls_relaxation.residue_name_2
      _nefpls_relaxation.atom_name_2

     1   1   .   1.000000   0.0170893   B   145   THR   H   B   145   THR   N

   stop_
"""


def test_t1noe_with_r1noe_data_single():
    """Test t1noe fitting with r1noe_data_single.nef test data."""

    test_data = open(path_in_test_data(__file__, "test_1_r1noe.nef")).read()

    # Test that function exits with error when no noise level provided
    result = run_and_report(app, ["T1_NOE_pos", "T1_NOE_neg"], input=test_data)

    r1_loop = isolate_loop(
        result.stdout, "nefpls_relaxation_list_r1", "nefpls_relaxation"
    )
    assert_lines_match(EXPECTED_R1_DATA_SINGLE_R1, r1_loop)

    noe_loop = isolate_loop(
        result.stdout, "nefpls_relaxation_list_noe", "nefpls_relaxation"
    )
    assert_lines_match(EXPECTED_R1_DATA_SINGLE_NOE, noe_loop)


def test_t1noe_with_r1noe_data_single_mc10_n0_1():
    """Test t1noe fitting with r1noe_data_single.nef test data."""

    test_data = open(path_in_test_data(__file__, "test_1_r1noe.nef")).read()

    # Test that function exits with error when no noise level provided
    result = run_and_report(
        app,
        ["T1_NOE_pos", "T1_NOE_neg", "--noise-level", "0.1", "--cycles", "10"],
        input=test_data,
    )

    r1_loop = isolate_loop(
        result.stdout, "nefpls_relaxation_list_r1", "nefpls_relaxation"
    )
    assert_lines_match(EXPECTED_R1_DATA_SINGLE_R1_MC, r1_loop)

    noe_loop = isolate_loop(
        result.stdout, "nefpls_relaxation_list_noe", "nefpls_relaxation"
    )
    assert_lines_match(EXPECTED_R1_DATA_SINGLE_NOE_MC, noe_loop)



import typer

from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()


if nef_app.app:

    nef_app.app.add_typer(
        app, name="nmrstar", help="- read NMR-STAR [sequences & shifts]"
    )

    app.add_typer(
        import_app,
        name="import",
        help="- import NMR-STAR sequences & shifts and whole projects",
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.nmrstar.importers.project_cli  # noqa: F401
    import nef_pipelines.transcoders.nmrstar.importers.rdcs  # noqa: F401
    import nef_pipelines.transcoders.nmrstar.importers.sequence  # noqa: F401
    import nef_pipelines.transcoders.nmrstar.importers.shifts  # noqa: F401



import pytest
import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.frames.tabulate import tabulate

runner = CliRunner()
app = typer.Typer()
app.command()(tabulate)


EXPECTED_BASIC = """
          nef_sequence
          ------------
      index  chain_code      sequence_code  residue_name    linking
         1            A             10               GLU    start
         2            A             11               TYR    middle
         3            A             12               ALA    middle
         4            A             13               GLN    middle
         5            A             14               PRO    middle
         6            A             15               ARG    middle
         7            A             16               LEU    middle
         8            A             17               ARG    middle
         9            A             18               LEU    middle
        10            A             19               GLY    middle
        11            A             20               PHE    middle
        12            A             21               GLU    middle
        13            A             22               ASP    end

    """

# noinspection PyUnusedLocal


def test_frame_basic():

    path = path_in_test_data(__file__, "tailin_seq_short.nef")

    result = run_and_report(app, ["--in", path])

    assert_lines_match(EXPECTED_BASIC, result.stdout)


def test_frame_selection():

    path = path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")

    result = run_and_report(app, ["--in", path, "nef_molecular_system"])

    assert "nef_sequence" in result.stdout
    assert "nef_chemical_shift_list_default" not in result.stdout


def test_frame_and_loop_selection_text():

    path = path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")

    result = run_and_report(
        app, ["--in", path, "nef_nmr_Spectrum_simnoe.nef_spectrum_Dimension"]
    )

    assert "nef_molecular_system" not in result.stdout
    assert "nef_spectrum_dimension" in result.stdout
    assert "nef_peak" not in result.stdout


def test_frame_and_loop_selection_index():

    path = path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")

    result = run_and_report(app, ["--in", path, "nef_nmr_spectrum_simnoe.3"])

    assert "nef_molecular_system" not in result.stdout
    assert "nef_spectrum_dimension" not in result.stdout
    assert "nef_peak" in result.stdout


def test_frame_and_loop_selection_text_exact():

    path = path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")

    result = run_and_report(
        app, ["--in", path, "nef_nmr_spectrum_simnoe.nef_spectrum_dimension", "--exact"]
    )

    assert "nef_molecular_system" not in result.stdout
    assert "nef_spectrum_dimension" in result.stdout
    assert "nef_peak" not in result.stdout


EXPECTED_SPECTRUM_DIMENSIONS_DIM_ONLY = """
    simnoe [nef_spectrum_dimension]
    -------------------------------

    dimension-id
               1
               2
               3
    """

EXPECTED_DIMENSION_ID_SPECTRAL_WIDTH = """
    simnoe [nef_spectrum_dimension]
    -------------------------------
      dimension-id    spectral-width
                 1          11.0113
                 2          11.06
                 3          81.8927
    """

EXPECTED_EXCLUDE_DIMENSION_ID_SPECTRAL_WIDTH = """
    simnoe [nef_spectrum_dimension]
    -------------------------------
      axis-unit    axis-code      spectrometer-frequency    folding    absolute-peak-positions
      ppm          1H                            800.133    circular   true
      ppm          1H                            800.133    circular   true
      ppm          15N                            81.076    circular   true

    """
EXPECTED_DATA_ALL = """  # noqa: E501
    simnoe [nef_spectrum_dimension]
    -------------------------------
      dimension-id  axis-unit    axis-code      spectrometer-frequency    spectral-width  folding    absolute-peak-positions
                 1  ppm          1H                            800.133           11.0113  circular   true
                 2  ppm          1H                            800.133           11.06    circular   true
                 3  ppm          15N                            81.076           81.8927  circular   true
    """.replace(
    NOQA_E501, ""
)

EXPECTED_DATA_LOOKUP = {
    "EXPECTED_DATA_EMPTY": "",
    "EXPECTED_SPECTRUM_DIMENSIONS_DIM_ONLY": EXPECTED_SPECTRUM_DIMENSIONS_DIM_ONLY,
    "EXPECTED_DIMENSION_ID_SPECTRAL_WIDTH": EXPECTED_DIMENSION_ID_SPECTRAL_WIDTH,
    "EXPECTED_EXCLUDE_DIMENSION_ID_SPECTRAL_WIDTH": EXPECTED_EXCLUDE_DIMENSION_ID_SPECTRAL_WIDTH,
    "EXPECTED_DATA_ALL": EXPECTED_DATA_ALL,
}


@pytest.mark.parametrize(
    "column_selections,expected_lookup",
    [
        ("-", "EXPECTED_DATA_EMPTY"),
        ("-*", "EXPECTED_DATA_EMPTY"),
        ("-,+dim", "EXPECTED_SPECTRUM_DIMENSIONS_DIM_ONLY"),
        ("-,dimension_id,spectral_width", "EXPECTED_DIMENSION_ID_SPECTRAL_WIDTH"),
        (
            "-dimension_id,-spectral_width",
            "EXPECTED_EXCLUDE_DIMENSION_ID_SPECTRAL_WIDTH",
        ),
        ("+", "EXPECTED_DATA_ALL"),
        ("+dimension_id,+spectral_width", "EXPECTED_DATA_ALL"),
    ],
)
def test_exclude_columns(column_selections, expected_lookup):

    path = path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")

    result = run_and_report(
        app,
        [
            "--in",
            path,
            "--abbreviate",
            "--select-columns",
            column_selections,
            "nef_nmr_spectrum_simnoe.nef_spectrum_dimension",
        ],
    )

    assert_lines_match(EXPECTED_DATA_LOOKUP[expected_lookup], result.stdout)


def test_full():
    path = path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")

    result = run_and_report(
        app, ["--in", path, "nef_nmr_spectrum_simnoe.nef_spectrum_dimension", "--full"]
    )

    assert "value_first_point" in result.stdout


def test_no_header():
    path = path_in_test_data(
        __file__,
        "tailin_seq_short.nef",
    )

    result = run_and_report(app, ["--in", path, "--no-title"])

    assert_lines_match("\n".join(EXPECTED_BASIC.split("\n")[3:]), result.stdout)


EXPECTED_CSV = """
index,chain_code,sequence_code,residue_name,linking
1,A,10,GLU,start
2,A,11,TYR,middle
3,A,12,ALA,middle
4,A,13,GLN,middle
5,A,14,PRO,middle
6,A,15,ARG,middle
7,A,16,LEU,middle
8,A,17,ARG,middle
9,A,18,LEU,middle
10,A,19,GLY,middle
11,A,20,PHE,middle
12,A,21,GLU,middle
13,A,22,ASP,end
"""


def test_csv():
    path = path_in_test_data(
        __file__,
        "tailin_seq_short.nef",
    )

    result = run_and_report(app, ["--in", path, "--no-title", "--format", "csv"])

    assert_lines_match(EXPECTED_CSV, result.stdout)
